# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helpers to apply enhancements."""
from __future__ import annotations

import os
from pathlib import Path

import yaml
from yaml import UnsafeLoader

from satpy._config import config_search_paths, get_entry_points_config_dirs
from satpy.decision_tree import DecisionTree
from satpy.utils import get_logger, recursive_dict_update

LOG = get_logger(__name__)


class EnhancementDecisionTree(DecisionTree):
    """The enhancement decision tree."""

    def __init__(self, *decision_dicts, **kwargs):
        """Init the decision tree."""
        match_keys = kwargs.pop("match_keys",
                                ("name",
                                 "reader",
                                 "platform_name",
                                 "sensor",
                                 "standard_name",
                                 "units",
                                 ))
        self.prefix = kwargs.pop("config_section", "enhancements")
        multival_keys = kwargs.pop("multival_keys", ["sensor"])
        super(EnhancementDecisionTree, self).__init__(
            decision_dicts, match_keys, multival_keys)

    def add_config_to_tree(self, *decision_dict: str | Path | dict) -> None:
        """Add configuration to tree."""
        conf: dict = {}
        for config_file in decision_dict:
            config_dict = self._get_config_dict_from_user(config_file)
            recursive_dict_update(conf, config_dict)
        self._build_tree(conf)

    def _get_config_dict_from_user(self, config_file: str | Path | dict) -> dict:
        if isinstance(config_file, (str, Path)) and os.path.isfile(config_file):
            config_dict = self._get_yaml_enhancement_dict(config_file)
        elif isinstance(config_file, dict):
            config_dict = config_file
        elif isinstance(config_file, str):
            LOG.debug("Loading enhancement config string")
            config_dict = yaml.load(config_file, Loader=UnsafeLoader)
            if not isinstance(config_dict, dict):
                raise ValueError(
                    "YAML file doesn't exist or string is not YAML dict: {}".format(config_file))
        else:
            raise ValueError(f"Unexpected type for enhancement configuration: {type(config_file)}")
        return config_dict

    def _get_yaml_enhancement_dict(self, config_file: str | Path) -> dict:
        with open(config_file) as fd:
            enhancement_config = yaml.load(fd, Loader=UnsafeLoader)
            if enhancement_config is None:
                # empty file
                return {}
            enhancement_section = enhancement_config.get(self.prefix, {})
            if not enhancement_section:
                LOG.debug("Config '{}' has no '{}' section or it is empty".format(config_file, self.prefix))
                return {}
            LOG.debug(f"Adding enhancement configuration from file: {config_file}")
        return enhancement_section

    def find_match(self, **query_dict):
        """Find a match."""
        try:
            return super(EnhancementDecisionTree, self).find_match(**query_dict)
        except KeyError:
            # give a more understandable error message
            raise KeyError("No enhancement configuration found for %s" %
                           (query_dict.get("uid", None),))


class Enhancer:
    """Helper class to get enhancement information for images."""

    def __init__(self, enhancement_config_file=None):
        """Initialize an Enhancer instance.

        Args:
            enhancement_config_file: The enhancement configuration to apply, False to leave as is.
        """
        self.enhancement_config_file = enhancement_config_file
        # Set enhancement_config_file to False for no enhancements
        if self.enhancement_config_file is None:
            # it wasn't specified in the config or in the kwargs, we should
            # provide a default
            config_fn = os.path.join("enhancements", "generic.yaml")
            paths = get_entry_points_config_dirs("satpy.enhancements")
            self.enhancement_config_file = config_search_paths(config_fn, search_dirs=paths)

        if not self.enhancement_config_file:
            # They don't want any automatic enhancements
            self.enhancement_tree = None
        else:
            if not isinstance(self.enhancement_config_file, (list, tuple)):
                self.enhancement_config_file = [self.enhancement_config_file]

            self.enhancement_tree = EnhancementDecisionTree(*self.enhancement_config_file)

        self.sensor_enhancement_configs = []

    def get_sensor_enhancement_config(self, sensor):
        """Get the sensor-specific config."""
        if isinstance(sensor, str):
            # one single sensor
            sensor = [sensor]

        paths = get_entry_points_config_dirs("satpy.enhancements")
        for sensor_name in sensor:
            config_fn = os.path.join("enhancements", sensor_name + ".yaml")
            config_files = config_search_paths(config_fn, search_dirs=paths)
            # Note: Enhancement configuration files can't overwrite individual
            # options, only entire sections are overwritten
            for config_file in config_files:
                yield config_file

    def add_sensor_enhancements(self, sensor):
        """Add sensor-specific enhancements."""
        # XXX: Should we just load all enhancements from the base directory?
        new_configs = []
        for config_file in self.get_sensor_enhancement_config(sensor):
            if config_file not in self.sensor_enhancement_configs:
                self.sensor_enhancement_configs.append(config_file)
                new_configs.append(config_file)

        if new_configs:
            self.enhancement_tree.add_config_to_tree(*new_configs)

    def apply(self, img, **info):
        """Apply the enhancements."""
        enh_kwargs = self.enhancement_tree.find_match(**info)

        backup_id = f"<name={info.get('name')}, calibration={info.get('calibration')}>"
        data_id = info.get("_satpy_id", backup_id)
        LOG.debug(f"Data for {data_id} will be enhanced with options:\n\t{enh_kwargs['operations']}")
        for operation in enh_kwargs["operations"]:
            fun = operation["method"]
            args = operation.get("args", [])
            kwargs = operation.get("kwargs", {})
            fun(img, *args, **kwargs)


def get_enhanced_image(dataset, enhance=None, overlay=None, decorate=None,
                       fill_value=None):
    """Get an enhanced version of `dataset` as an :class:`~trollimage.xrimage.XRImage` instance.

    Args:
        dataset (xarray.DataArray): Data to be enhanced and converted to an image.
        enhance (bool or satpy.enhancements.enhancer.Enhancer): Whether to automatically enhance
            data to be more visually useful and to fit inside the file
            format being saved to. By default, this will default to using
            the enhancement configuration files found using the default
            :class:`~satpy.enhancements.enhancer.Enhancer` class. This can be set to
            `False` so that no enhancments are performed. This can also
            be an instance of the :class:`~satpy.enhancements.enhancer.Enhancer` class
            if further custom enhancement is needed.
        overlay (dict): Options for image overlays. See :func:`~satpy.enhancements.overlays.add_overlay`
            for available options.
        decorate (dict): Options for decorating the image. See
            :func:`~satpy.enhancements.overlays.add_decorate` for available options.
        fill_value (int or float): Value to use when pixels are masked or
            invalid. Default of `None` means to create an alpha channel.
            See :meth:`~trollimage.xrimage.XRImage.finalize` for more
            details. Only used when adding overlays or decorations. Otherwise
            it is up to the caller to "finalize" the image before using it
            except if calling ``img.show()`` or providing the image to
            a writer as these will finalize the image.
    """
    from trollimage.xrimage import XRImage

    if enhance is False:
        # no enhancement
        enhancer = None
    elif enhance is None or enhance is True:
        # default enhancement
        from satpy.enhancements.enhancer import Enhancer

        enhancer = Enhancer()
    else:
        # custom enhancer
        enhancer = enhance

    # Create an image for enhancement
    img = XRImage(dataset)

    if enhancer is None or enhancer.enhancement_tree is None:
        LOG.debug("No enhancement being applied to dataset")
    else:
        if dataset.attrs.get("sensor", None):
            enhancer.add_sensor_enhancements(dataset.attrs["sensor"])

        enhancer.apply(img, **dataset.attrs)

    if overlay is not None:
        from satpy.enhancements.overlays import add_overlay

        img = add_overlay(img, dataset.attrs["area"], fill_value=fill_value, **overlay)

    if decorate is not None:
        from satpy.enhancements.overlays import add_decorate

        img = add_decorate(img, fill_value=fill_value, **decorate)

    return img



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.acspo module."""

from __future__ import annotations

import datetime as dt
import os
from typing import Any
from unittest import mock

import numpy as np
import pytest

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_SHAPE = (10, 300)


class FakeNetCDF4FileHandler2(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename: str, filename_info: dict, filetype_info: dict) -> dict:
        """Mimic reader input file content."""
        DEFAULT_FILE_DTYPE = np.uint16
        DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                                      dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
        DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
        DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
        DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
        DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)

        date = filename_info.get("start_time", dt.datetime(2016, 1, 1, 12, 0, 0))
        sat, inst = {
            "VIIRS_NPP": ("NPP", "VIIRS"),
            "VIIRS_N20": ("N20", "VIIRS"),
        }[filename_info["sensor_id"]]

        file_content: dict[str, Any] = {
            "/attr/platform": sat,
            "/attr/sensor": inst,
            "/attr/spatial_resolution": "742 m at nadir",
            "/attr/time_coverage_start": date.strftime("%Y%m%dT%H%M%SZ"),
            "/attr/time_coverage_end": (date + dt.timedelta(minutes=6)).strftime("%Y%m%dT%H%M%SZ"),
        }

        file_content["lat"] = DEFAULT_LAT_DATA
        file_content["lat/attr/comment"] = "Latitude of retrievals"
        file_content["lat/attr/long_name"] = "latitude"
        file_content["lat/attr/standard_name"] = "latitude"
        file_content["lat/attr/units"] = "degrees_north"
        file_content["lat/attr/valid_min"] = -90.
        file_content["lat/attr/valid_max"] = 90.
        file_content["lat/shape"] = DEFAULT_FILE_SHAPE

        file_content["lon"] = DEFAULT_LON_DATA
        file_content["lon/attr/comment"] = "Longitude of retrievals"
        file_content["lon/attr/long_name"] = "longitude"
        file_content["lon/attr/standard_name"] = "longitude"
        file_content["lon/attr/units"] = "degrees_east"
        file_content["lon/attr/valid_min"] = -180.
        file_content["lon/attr/valid_max"] = 180.
        file_content["lon/shape"] = DEFAULT_FILE_SHAPE

        for k in ["sea_surface_temperature",
                  "satellite_zenith_angle",
                  "sea_ice_fraction",
                  "wind_speed"]:
            file_content[k] = DEFAULT_FILE_DATA[None, ...]
            file_content[k + "/attr/scale_factor"] = 1.1
            file_content[k + "/attr/add_offset"] = 0.1
            file_content[k + "/attr/units"] = "some_units"
            file_content[k + "/attr/comment"] = "comment"
            file_content[k + "/attr/standard_name"] = "standard_name"
            file_content[k + "/attr/long_name"] = "long_name"
            file_content[k + "/attr/valid_min"] = 0
            file_content[k + "/attr/valid_max"] = 65534
            file_content[k + "/attr/_FillValue"] = 65534
            file_content[k + "/shape"] = (1, DEFAULT_FILE_SHAPE[0], DEFAULT_FILE_SHAPE[1])

        l2p_flags_data = np.zeros(
            (1, DEFAULT_FILE_SHAPE[0], DEFAULT_FILE_SHAPE[1]),
            dtype=np.int16)
        l2p_flags_data[0, -1, -1] = np.int16(np.uint16(0b1100000000000000))  # cloud
        file_content["l2p_flags"] = l2p_flags_data

        convert_file_content_to_data_array(file_content, dims=("time", "nj", "ni"))
        return file_content


class TestACSPOReader:
    """Test ACSPO Reader."""

    yaml_file = "acspo.yaml"

    def setup_method(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.acspo import ACSPOFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(ACSPOFileHandler, "__bases__", (FakeNetCDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def teardown_method(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    @pytest.mark.parametrize(
        "filename",
        [
            ("20170401174600-STAR-L2P_GHRSST-SSTskin-VIIRS_NPP-ACSPO_V2.40-v02.0-fv01.0.nc"),
            ("20210916161708-STAR-L2P_GHRSST-SSTsubskin-VIIRS_N20-ACSPO_V2.80-v02.0-fv01.0.nc"),
        ]
    )
    def test_init(self, filename):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([filename])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    @pytest.mark.parametrize(
        ("var_name", "cloud_clearable", "cloud_clear"),
        [
            ("sst", True, False),
            ("sst", True, True),
            ("satellite_zenith_angle", False, False),
            ("satellite_zenith_angle", False, True),
            ("sea_ice_fraction", False, False),
            ("wind_speed", False, False),
        ])
    def test_load_every_dataset(self, var_name, cloud_clearable, cloud_clear):
        """Test loading all datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "20170401174600-STAR-L2P_GHRSST-SSTskin-VIIRS_NPP-ACSPO_V2.40-v02.0-fv01.0.nc",
        ])
        fh_kwargs = {}
        if cloud_clear:
            fh_kwargs["filters"] = ["cloud_clear"]
        r.create_filehandlers(loadables, fh_kwargs=fh_kwargs)
        datasets = r.load([var_name])
        assert len(datasets) == 1
        d = datasets[var_name]
        assert d.shape == DEFAULT_FILE_SHAPE
        assert d.dims == ("y", "x")
        assert d.attrs["sensor"] == "viirs"
        assert d.attrs["rows_per_scan"] == 16
        dask_data = d.data
        np_data = dask_data.compute()
        assert dask_data.dtype == np.float32
        assert np_data.dtype == np.float32

        exp_data = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1], dtype=np.uint16).reshape(DEFAULT_FILE_SHAPE)
        exp_data = exp_data * np.float32(1.1) + 0.1
        if cloud_clearable and cloud_clear:
            exp_data[-1, -1] = np.nan
        np.testing.assert_array_equal(np_data, exp_data, strict=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The ahi_hsd reader tests package."""

from __future__ import annotations

import contextlib
import datetime as dt
import unittest
import warnings
from typing import Any, Dict
from unittest import mock

import dask.array as da
import numpy as np
import pytest

from satpy.readers.ahi_hsd import AHIHSDFileHandler, _NominalTimeCalculator
from satpy.readers.core.utils import get_geostationary_mask
from satpy.tests.utils import make_dataid

InfoDict = Dict[str, Any]

FAKE_BASIC_INFO: InfoDict = {
    "blocklength": 0,
    "satellite": "Himawari-8",
    "observation_area": "FLDK",
    "observation_start_time": 58413.12523839,
    "observation_end_time": 58413.132182834444,
    "observation_timeline": "0300",
}
FAKE_DATA_INFO: InfoDict = {
    "blocklength": 50,
    "compression_flag_for_data": 0,
    "hblock_number": 2,
    "number_of_bits_per_pixel": 16,
    "number_of_columns": np.array([11000]),
    "number_of_lines": np.array([1100]),
    "spare": "",
}
FAKE_PROJ_INFO: InfoDict = {
    "CFAC": 40932549,
    "COFF": 5500.5,
    "LFAC": 40932549,
    "LOFF": 5500.5,
    "blocklength": 127,
    "coeff_for_sd": 1737122264.0,
    "distance_from_earth_center": 42164.0,
    "earth_equatorial_radius": 6378.137,
    "earth_polar_radius": 6356.7523,
    "hblock_number": 3,
    "req2_rpol2": 1.006739501,
    "req2_rpol2_req2": 0.0066943844,
    "resampling_size": 4,
    "resampling_types": 0,
    "rpol2_req2": 0.993305616,
    "spare": "",
    "sub_lon": 140.7,
}
FAKE_NAV_INFO: InfoDict = {
    "SSP_longitude": 140.65699999999998,
    "SSP_latitude": 0.0042985719753897015,
    "distance_earth_center_to_satellite": 42165.04,
    "nadir_longitude": 140.25253875463318,
    "nadir_latitude": 0.01674775121155575,
}
FAKE_CAL_INFO: InfoDict = {"blocklength": 0, "band_number": [4]}
FAKE_IRVISCAL_INFO: InfoDict = {}
FAKE_INTERCAL_INFO: InfoDict = {"blocklength": 0}
FAKE_SEGMENT_INFO: InfoDict = {"blocklength": 0}
FAKE_NAVCORR_INFO: InfoDict = {"blocklength": 0, "numof_correction_info_data": [1]}
FAKE_NAVCORR_SUBINFO: InfoDict = {}
FAKE_OBS_TIME_INFO: InfoDict = {"blocklength": 0, "number_of_observation_times": [1]}
FAKE_OBS_LINETIME_INFO: InfoDict = {}
FAKE_ERROR_INFO: InfoDict = {"blocklength": 0, "number_of_error_info_data": [1]}
FAKE_ERROR_LINE_INFO: InfoDict = {}
FAKE_SPARE_INFO: InfoDict = {"blocklength": 0}


def _new_unzip(fname, prefix=""):
    """Fake unzipping."""
    if fname[-3:] == "bz2":
        return prefix + fname[:-4]
    return fname


class TestAHIHSDNavigation(unittest.TestCase):
    """Test the AHI HSD reader navigation."""

    @mock.patch("satpy.readers.ahi_hsd.np2str")
    @mock.patch("satpy.readers.ahi_hsd.np.fromfile")
    def test_region(self, fromfile, np2str):
        """Test region navigation."""
        from pyproj import CRS

        np2str.side_effect = lambda x: x
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_hsd.open", m, create=True):
            fh = AHIHSDFileHandler("somefile",
                                   {"segment": 1, "total_segments": 1},
                                   filetype_info={"file_type": "hsd_b01"},
                                   user_calibration=None)
            fh.proj_info = {"CFAC": 40932549,
                            "COFF": -591.5,
                            "LFAC": 40932549,
                            "LOFF": 5132.5,
                            "blocklength": 127,
                            "coeff_for_sd": 1737122264.0,
                            "distance_from_earth_center": 42164.0,
                            "earth_equatorial_radius": 6378.137,
                            "earth_polar_radius": 6356.7523,
                            "hblock_number": 3,
                            "req2_rpol2": 1.006739501,
                            "req2_rpol2_req2": 0.0066943844,
                            "resampling_size": 4,
                            "resampling_types": 0,
                            "rpol2_req2": 0.993305616,
                            "spare": "",
                            "sub_lon": 140.7}

            fh.data_info = {"blocklength": 50,
                            "compression_flag_for_data": 0,
                            "hblock_number": 2,
                            "number_of_bits_per_pixel": 16,
                            "number_of_columns": np.array([1000]),
                            "number_of_lines": np.array([1000]),
                            "spare": ""}

            area_def = fh.get_area_def(None)
            expected_crs = CRS.from_dict(dict(a=6378137.0, b=6356752.3, h= 35785863.0,
                                              lon_0=140.7, proj="geos", units="m"))
            assert area_def.crs == expected_crs
            np.testing.assert_allclose(area_def.area_extent, (592000.0038256242, 4132000.0267018233,
                                                              1592000.0102878273, 5132000.033164027))

    @mock.patch("satpy.readers.ahi_hsd.np2str")
    @mock.patch("satpy.readers.ahi_hsd.np.fromfile")
    def test_segment(self, fromfile, np2str):
        """Test segment navigation."""
        from pyproj import CRS

        np2str.side_effect = lambda x: x
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_hsd.open", m, create=True):
            fh = AHIHSDFileHandler("somefile", {"segment": 8, "total_segments": 10},
                                   filetype_info={"file_type": "hsd_b01"})
            fh.proj_info = {"CFAC": 40932549,
                            "COFF": 5500.5,
                            "LFAC": 40932549,
                            "LOFF": 5500.5,
                            "blocklength": 127,
                            "coeff_for_sd": 1737122264.0,
                            "distance_from_earth_center": 42164.0,
                            "earth_equatorial_radius": 6378.137,
                            "earth_polar_radius": 6356.7523,
                            "hblock_number": 3,
                            "req2_rpol2": 1.006739501,
                            "req2_rpol2_req2": 0.0066943844,
                            "resampling_size": 4,
                            "resampling_types": 0,
                            "rpol2_req2": 0.993305616,
                            "spare": "",
                            "sub_lon": 140.7}

            fh.data_info = {"blocklength": 50,
                            "compression_flag_for_data": 0,
                            "hblock_number": 2,
                            "number_of_bits_per_pixel": 16,
                            "number_of_columns": np.array([11000]),
                            "number_of_lines": np.array([1100]),
                            "spare": ""}

            area_def = fh.get_area_def(None)
            expected_crs = CRS.from_dict(dict(a=6378137.0, b=6356752.3, h= 35785863.0,
                                              lon_0=140.7, proj="geos", units="m"))
            assert area_def.crs == expected_crs
            np.testing.assert_allclose(area_def.area_extent, (-5500000.035542117, -3300000.021325271,
                                                              5500000.035542117, -2200000.0142168473))


@pytest.fixture
def hsd_file_jp01(tmp_path):
    """Create a jp01 hsd file."""
    from satpy.readers.ahi_hsd import (  # _IRCAL_INFO_TYPE,
        _BASIC_INFO_TYPE,
        _CAL_INFO_TYPE,
        _DATA_INFO_TYPE,
        _ERROR_INFO_TYPE,
        _ERROR_LINE_INFO_TYPE,
        _INTER_CALIBRATION_INFO_TYPE,
        _NAV_INFO_TYPE,
        _NAVIGATION_CORRECTION_INFO_TYPE,
        _NAVIGATION_CORRECTION_SUBINFO_TYPE,
        _OBSERVATION_LINE_TIME_INFO_TYPE,
        _OBSERVATION_TIME_INFO_TYPE,
        _PROJ_INFO_TYPE,
        _SEGMENT_INFO_TYPE,
        _SPARE_TYPE,
        _VISCAL_INFO_TYPE,
    )
    nrows = 11000
    ncols = 11000
    filename = tmp_path / "somedata.DAT"
    error_lines = 0
    number_nav_corrections = 0
    number_observation_times = 6
    dat_type = np.dtype([("block1", _BASIC_INFO_TYPE),
                         ("block2", _DATA_INFO_TYPE),
                         ("block3", _PROJ_INFO_TYPE),
                         ("block4", _NAV_INFO_TYPE),
                         ("block5", _CAL_INFO_TYPE),
                         ("calibration", _VISCAL_INFO_TYPE),
                         ("block6", _INTER_CALIBRATION_INFO_TYPE),
                         ("block7", _SEGMENT_INFO_TYPE),
                         ("block8", _NAVIGATION_CORRECTION_INFO_TYPE),
                         ("navigation_corrections", _NAVIGATION_CORRECTION_SUBINFO_TYPE,
                          (number_nav_corrections,)),
                         ("block9", _OBSERVATION_TIME_INFO_TYPE),
                         ("observation_time_information", _OBSERVATION_LINE_TIME_INFO_TYPE,
                          (number_observation_times,)),
                         ("block10", _ERROR_INFO_TYPE),
                         ("error_info", _ERROR_LINE_INFO_TYPE, (error_lines,)),
                         ("block11", _SPARE_TYPE),
                         ("image", "<u2", (nrows, ncols))])
    dat = np.zeros(1, dat_type)
    dat["block1"]["blocklength"] = _BASIC_INFO_TYPE.itemsize
    dat["block1"]["observation_area"] = "JP01"
    dat["block1"]["satellite"] = b"Himawari-8"
    dat["block2"]["blocklength"] = _DATA_INFO_TYPE.itemsize
    dat["block2"]["number_of_lines"] = nrows
    dat["block2"]["number_of_columns"] = ncols
    dat["block3"]["blocklength"] = _PROJ_INFO_TYPE.itemsize
    dat["block3"]["sub_lon"] = 140.7
    dat["block3"]["CFAC"] = 81865099
    dat["block3"]["LFAC"] = 81865099
    dat["block3"]["COFF"] = 5500.5
    dat["block3"]["LOFF"] = 5500.5
    dat["block3"]["distance_from_earth_center"] = 42164.
    dat["block3"]["earth_equatorial_radius"] = 6378.137
    dat["block3"]["earth_polar_radius"] = 6356.7523
    dat["block4"]["blocklength"] = _NAV_INFO_TYPE.itemsize
    dat["block5"]["blocklength"] = _CAL_INFO_TYPE.itemsize + _VISCAL_INFO_TYPE.itemsize
    dat["block6"]["blocklength"] = _INTER_CALIBRATION_INFO_TYPE.itemsize
    dat["block7"]["blocklength"] = _SEGMENT_INFO_TYPE.itemsize
    dat["block8"]["blocklength"] = (_NAVIGATION_CORRECTION_INFO_TYPE.itemsize +
                                    number_nav_corrections * _NAVIGATION_CORRECTION_SUBINFO_TYPE.itemsize)
    dat["block9"]["blocklength"] = (_OBSERVATION_TIME_INFO_TYPE.itemsize +
                                    number_observation_times * _OBSERVATION_LINE_TIME_INFO_TYPE.itemsize)
    dat["block10"]["blocklength"] = _ERROR_INFO_TYPE.itemsize + error_lines * _ERROR_LINE_INFO_TYPE.itemsize
    dat["block11"]["blocklength"] = _SPARE_TYPE.itemsize
    dat.tofile(filename)
    return filename


class TestAHIHSDFileHandler:
    """Tests for the AHI HSD file handler."""

    def test_bad_calibration(self):
        """Test that a bad calibration mode causes an exception."""
        with pytest.raises(ValueError, match="Invalid calibration mode: BAD_MODE. Choose one of (.*)"):
            with _fake_hsd_handler(fh_kwargs={"calib_mode": "BAD_MODE"}):
                pass

    @pytest.mark.parametrize(
        ("round_actual_position", "expected_result"),
        [
            (False, (140.65699999999998, 0.0042985719753897015, 35786903.00011936)),
            (True, (140.657, 0.0, 35786850.0))
        ]
    )
    def test_actual_satellite_position(self, round_actual_position, expected_result):
        """Test that rounding of the actual satellite position can be controlled."""
        with _fake_hsd_handler(fh_kwargs={"round_actual_position": round_actual_position}) as fh:
            ds_id = make_dataid(name="B01", resolution=1000)
            ds_info = {
                "units": "%",
                "standard_name": "some_name",
                "wavelength": (0.1, 0.2, 0.3),
                "resolution": 1000,
            }
            metadata = fh._get_metadata(ds_id, ds_info)
            orb_params = metadata["orbital_parameters"]
            assert orb_params["satellite_actual_longitude"] == expected_result[0]
            assert orb_params["satellite_actual_latitude"] == expected_result[1]
            assert orb_params["satellite_actual_altitude"] == expected_result[2]

    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler._check_fpos")
    def test_read_header(self, *mocks):
        """Test header reading."""
        with _fake_hsd_handler() as fh:
            fh._read_header(mock.MagicMock())

    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler._read_data")
    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_invalid")
    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler.calibrate")
    def test_read_band(self, calibrate, *mocks):
        """Test masking of space pixels."""
        nrows = 25
        ncols = 100
        calibrate.return_value = np.ones((nrows, ncols))
        with _fake_hsd_handler() as fh:
            fh.data_info["number_of_columns"] = ncols
            fh.data_info["number_of_lines"] = nrows
            with warnings.catch_warnings():
                # The header isn't valid
                warnings.filterwarnings("ignore", category=UserWarning, message=r"Actual .* header size")
                im = fh.read_band(mock.MagicMock(), mock.MagicMock())
            # Note: Within the earth's shape get_geostationary_mask() is True but the numpy.ma mask
            # is False
            mask = im.to_masked_array().mask
            ref_mask = np.logical_not(get_geostationary_mask(fh.area).compute())
            np.testing.assert_equal(mask, ref_mask)

            # Test attributes
            orb_params_exp = {"projection_longitude": 140.7,
                              "projection_latitude": 0.,
                              "projection_altitude": 35785863.0,
                              "satellite_actual_longitude": 140.657,
                              "satellite_actual_latitude": 0.0,
                              "satellite_actual_altitude": 35786850,
                              "nadir_longitude": 140.252539,
                              "nadir_latitude": 0.01674775}
            actual_obs_params = im.attrs["orbital_parameters"]
            for key, value in orb_params_exp.items():
                assert key in actual_obs_params
                np.testing.assert_allclose(value, actual_obs_params[key])

            time_params_exp = {
                "nominal_start_time": dt.datetime(2018, 10, 22, 3, 0, 0, 0),
                "nominal_end_time": dt.datetime(2018, 10, 22, 3, 10, 0, 0),
                "observation_start_time": dt.datetime(2018, 10, 22, 3, 0, 20, 596896),
                "observation_end_time": dt.datetime(2018, 10, 22, 3, 10, 20, 596896),
            }
            actual_time_params = im.attrs["time_parameters"]
            for key, value in time_params_exp.items():
                assert key in actual_time_params
                assert value == actual_time_params[key]

            # Test if masking space pixels disables with appropriate flag
            fh.mask_space = False
            with mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_space") as mask_space:
                with warnings.catch_warnings():
                    # The header isn't valid
                    warnings.filterwarnings("ignore", category=UserWarning, message=r"Actual .* header size")
                    fh.read_band(mock.MagicMock(), mock.MagicMock())
                mask_space.assert_not_called()

    def test_read_band_from_actual_file(self, hsd_file_jp01):
        """Test read_bands on real data."""
        filename_info = {"segment": 1, "total_segments": 1}
        filetype_info = {"file_type": "blahB01"}
        fh = AHIHSDFileHandler(hsd_file_jp01, filename_info, filetype_info)
        key = {"name": "B01", "calibration": "counts", "resolution": 1000}
        import dask
        with dask.config.set({"array.chunk-size": "32MiB"}):
            with warnings.catch_warnings():
                # The header isn't valid
                warnings.filterwarnings("ignore", category=UserWarning, message=r"Actual .* header size")
                data = fh.read_band(
                    key,
                    {
                        "units": "%",
                        "standard_name": "toa_bidirectional_reflectance",
                        "wavelength": 2,
                        "resolution": 1000,
                    })
            assert data.chunks == ((1100,) * 10, (1100,) * 10)
            assert data.dtype == data.compute().dtype
            assert data.dtype == np.float32

    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler._read_data")
    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_invalid")
    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler.calibrate")
    def test_scene_loading(self, calibrate, *mocks):
        """Test masking of space pixels."""
        from satpy import Scene
        nrows = 25
        ncols = 100
        calibrate.return_value = np.ones((nrows, ncols))
        with _fake_hsd_handler() as fh:
            with mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler") as fh_cls:
                fh_cls.return_value = fh
                fh.filename_info["total_segments"] = 1
                fh.filename_info["segment"] = 1
                fh.data_info["number_of_columns"] = ncols
                fh.data_info["number_of_lines"] = nrows
                scn = Scene(reader="ahi_hsd", filenames=["HS_H08_20210225_0700_B07_FLDK_R20_S0110.DAT"])
                with warnings.catch_warnings():
                    # The header isn't valid
                    warnings.filterwarnings("ignore", category=UserWarning, message=r"Actual .* header size")
                    scn.load(["B07"])
                im = scn["B07"]

                # Make sure space masking worked
                mask = im.to_masked_array().mask
                ref_mask = np.logical_not(get_geostationary_mask(fh.area).compute())
                np.testing.assert_equal(mask, ref_mask)

                # Make sure proj_id is correct
                assert fh.area.proj_id == f'geosh{FAKE_BASIC_INFO["satellite"][-1]}'

    def test_time_properties(self):
        """Test start/end/scheduled time properties."""
        with _fake_hsd_handler() as fh:
            assert fh.start_time == dt.datetime(2018, 10, 22, 3, 0)
            assert fh.end_time == dt.datetime(2018, 10, 22, 3, 10)
            assert fh.observation_start_time == dt.datetime(2018, 10, 22, 3, 0, 20, 596896)
            assert fh.observation_end_time == dt.datetime(2018, 10, 22, 3, 10, 20, 596896)
            assert fh.nominal_start_time == dt.datetime(2018, 10, 22, 3, 0, 0, 0)
            assert fh.nominal_end_time == dt.datetime(2018, 10, 22, 3, 10, 0, 0)

    def test_blocklen_error(self, *mocks):
        """Test erraneous blocklength."""
        open_name = "%s.open" % __name__
        fpos = 50
        with _fake_hsd_handler() as fh, \
                mock.patch(open_name, create=True) as mock_open, \
                mock_open(mock.MagicMock(), "r") as fp_:
            # Expected and actual blocklength match
            fp_.tell.return_value = 50
            with warnings.catch_warnings(record=True) as w:
                fh._check_fpos(fp_, fpos, 0, "header 1")
                assert len(w) == 0

            # Expected and actual blocklength do not match
            fp_.tell.return_value = 100
            with pytest.warns(UserWarning, match=r"Actual .* header size does not match expected"):
                fh._check_fpos(fp_, fpos, 0, "header 1")


class TestAHICalibration(unittest.TestCase):
    """Test case for various AHI calibration types."""

    @mock.patch("satpy.readers.ahi_hsd.AHIHSDFileHandler.__init__",
                return_value=None)
    def setUp(self, *mocks):
        """Create fake data for testing."""
        self.def_cali = [-0.0037, 15.20]
        self.upd_cali = [-0.0074, 30.40]
        self.bad_cali = [0.0, 0.0]
        fh = AHIHSDFileHandler(filetype_info={"file_type": "hsd_b01"})
        fh.calib_mode = "NOMINAL"
        fh.user_calibration = None
        fh.is_zipped = False
        fh._header = {
            "block5": {"band_number": [5],
                       "gain_count2rad_conversion": [self.def_cali[0]],
                       "offset_count2rad_conversion": [self.def_cali[1]],
                       "central_wave_length": [10.4073], },
            "calibration": {"coeff_rad2albedo_conversion": [0.0019255],
                            "speed_of_light": [299792458.0],
                            "planck_constant": [6.62606957e-34],
                            "boltzmann_constant": [1.3806488e-23],
                            "c0_rad2tb_conversion": [-0.116127314574],
                            "c1_rad2tb_conversion": [1.00099153832],
                            "c2_rad2tb_conversion": [-1.76961091571e-06],
                            "cali_gain_count2rad_conversion": [self.upd_cali[0]],
                            "cali_offset_count2rad_conversion": [self.upd_cali[1]]},
        }

        self.counts = da.array(np.array([[0, 1000],
                                         [2000, 5000]], dtype=np.uint16))
        self.fh = fh

    def test_default_calibrate(self, *mocks):
        """Test default in-file calibration modes."""
        self.setUp()
        # Counts
        assert self.fh.calibrate(data=123, calibration="counts") == 123

        # Radiance
        rad_exp = np.array([[15.2, 11.5],
                            [7.8, -3.3]])
        rad = self.fh.calibrate(data=self.counts,
                                calibration="radiance")
        assert np.allclose(rad, rad_exp)

        # Brightness Temperature
        bt_exp = np.array([[330.978979, 310.524688],
                           [285.845017, np.nan]])
        bt = self.fh.calibrate(data=self.counts,
                               calibration="brightness_temperature")
        np.testing.assert_allclose(bt, bt_exp)

        # Reflectance
        refl_exp = np.array([[2.92676, 2.214325],
                             [1.50189, 0.]])
        refl = self.fh.calibrate(data=self.counts,
                                 calibration="reflectance")
        assert np.allclose(refl, refl_exp)

    def test_updated_calibrate(self):
        """Test updated in-file calibration modes."""
        # Standard operation
        self.fh.calib_mode = "UPDATE"
        rad_exp = np.array([[30.4, 23.0],
                            [15.6, -6.6]])
        rad = self.fh.calibrate(data=self.counts, calibration="radiance")
        assert np.allclose(rad, rad_exp)

        # Case for no updated calibration available (older data)
        self.fh._header = {
            "block5": {"band_number": [5],
                       "gain_count2rad_conversion": [self.def_cali[0]],
                       "offset_count2rad_conversion": [self.def_cali[1]],
                       "central_wave_length": [10.4073], },
            "calibration": {"coeff_rad2albedo_conversion": [0.0019255],
                            "speed_of_light": [299792458.0],
                            "planck_constant": [6.62606957e-34],
                            "boltzmann_constant": [1.3806488e-23],
                            "c0_rad2tb_conversion": [-0.116127314574],
                            "c1_rad2tb_conversion": [1.00099153832],
                            "c2_rad2tb_conversion": [-1.76961091571e-06],
                            "cali_gain_count2rad_conversion": [self.bad_cali[0]],
                            "cali_offset_count2rad_conversion": [self.bad_cali[1]]},
        }
        rad = self.fh.calibrate(data=self.counts, calibration="radiance")
        rad_exp = np.array([[15.2, 11.5],
                            [7.8, -3.3]])
        assert np.allclose(rad, rad_exp)

    def test_user_calibration(self):
        """Test user-defined calibration modes."""
        # This is for radiance correction
        self.fh.user_calibration = {"B13": {"slope": 0.95,
                                            "offset": -0.1}}
        self.fh.band_name = "B13"
        rad = self.fh.calibrate(data=self.counts, calibration="radiance")
        rad_np = rad.compute()
        assert rad.dtype == rad_np.dtype
        assert rad.dtype == np.float32
        rad_exp = np.array([[16.10526316, 12.21052632],
                            [8.31578947, -3.36842105]])
        assert np.allclose(rad, rad_exp)

        # This is for DN calibration
        self.fh.user_calibration = {"B13": {"slope": -0.0032,
                                            "offset": 15.20},
                                    "type": "DN"}
        self.fh.band_name = "B13"
        rad = self.fh.calibrate(data=self.counts, calibration="radiance")
        rad_np = rad.compute()
        assert rad.dtype == rad_np.dtype
        assert rad.dtype == np.float32
        rad_exp = np.array([[15.2, 12.],
                            [8.8, -0.8]])
        assert np.allclose(rad, rad_exp)


@contextlib.contextmanager
def _fake_hsd_handler(fh_kwargs=None):
    """Create a test file handler."""
    m = mock.mock_open()
    with mock.patch("satpy.readers.ahi_hsd.np.fromfile", _custom_fromfile), \
            mock.patch("satpy.readers.ahi_hsd.unzip_file", mock.MagicMock(side_effect=_new_unzip)), \
            mock.patch("satpy.readers.ahi_hsd.open", m, create=True):
        in_fname = "test_file.bz2"
        fh = _create_fake_file_handler(in_fname, fh_kwargs=fh_kwargs)
        yield fh


def _custom_fromfile(*args, **kwargs):
    from satpy.readers.ahi_hsd import (
        _BASIC_INFO_TYPE,
        _CAL_INFO_TYPE,
        _DATA_INFO_TYPE,
        _ERROR_INFO_TYPE,
        _ERROR_LINE_INFO_TYPE,
        _INTER_CALIBRATION_INFO_TYPE,
        _IRCAL_INFO_TYPE,
        _NAV_INFO_TYPE,
        _NAVIGATION_CORRECTION_INFO_TYPE,
        _NAVIGATION_CORRECTION_SUBINFO_TYPE,
        _OBSERVATION_LINE_TIME_INFO_TYPE,
        _OBSERVATION_TIME_INFO_TYPE,
        _PROJ_INFO_TYPE,
        _SEGMENT_INFO_TYPE,
        _SPARE_TYPE,
        _VISCAL_INFO_TYPE,
    )
    dtype = kwargs.get("dtype")
    fake_info_map = {
        _BASIC_INFO_TYPE: FAKE_BASIC_INFO,
        _DATA_INFO_TYPE: FAKE_DATA_INFO,
        _NAV_INFO_TYPE: FAKE_NAV_INFO,
        _PROJ_INFO_TYPE: FAKE_PROJ_INFO,
        _CAL_INFO_TYPE: FAKE_CAL_INFO,
        _VISCAL_INFO_TYPE: FAKE_IRVISCAL_INFO,
        _IRCAL_INFO_TYPE: FAKE_IRVISCAL_INFO,
        _INTER_CALIBRATION_INFO_TYPE: FAKE_INTERCAL_INFO,
        _SEGMENT_INFO_TYPE: FAKE_SEGMENT_INFO,
        _NAVIGATION_CORRECTION_INFO_TYPE: FAKE_NAVCORR_INFO,
        _NAVIGATION_CORRECTION_SUBINFO_TYPE: FAKE_NAVCORR_SUBINFO,
        _OBSERVATION_TIME_INFO_TYPE: FAKE_OBS_TIME_INFO,
        _OBSERVATION_LINE_TIME_INFO_TYPE: FAKE_OBS_LINETIME_INFO,
        _ERROR_INFO_TYPE: FAKE_ERROR_INFO,
        _ERROR_LINE_INFO_TYPE: FAKE_ERROR_LINE_INFO,
        _SPARE_TYPE: FAKE_SPARE_INFO,
    }
    info_dict = fake_info_map[dtype]
    fake_arr = np.zeros((1,), dtype=dtype)
    for key, value in info_dict.items():
        fake_arr[key] = value
    return fake_arr


def _create_fake_file_handler(in_fname, filename_info=None, filetype_info=None, fh_kwargs=None):
    if filename_info is None:
        filename_info = {"segment": 8, "total_segments": 10}
    if filetype_info is None:
        filetype_info = {"file_type": "hsd_b01"}
    if fh_kwargs is None:
        fh_kwargs = {}
    fh = AHIHSDFileHandler(in_fname, filename_info, filetype_info, **fh_kwargs)

    # Check that the filename is altered and 2 digit segment prefix added for bz2 format files
    assert in_fname != fh.filename
    assert str(filename_info["segment"]).zfill(2) == fh.filename[0:2]
    return fh


class TestNominalTimeCalculator:
    """Test case for nominal timestamp computation."""

    @pytest.mark.parametrize(
        ("timeline", "expected"),
        [
            ("0300", dt.datetime(2020, 1, 1, 3, 0, 0)),
            ("65526", dt.datetime(2020, 1, 1, 12, 0, 0))
        ]
    )
    def test_invalid_timeline(self, timeline, expected):
        """Test handling of invalid timeline."""
        calc = _NominalTimeCalculator(timeline, "FLDK")
        exp_warning = contextlib.nullcontext()
        if timeline == "65526":
            exp_warning = pytest.warns(UserWarning, match="Observation timeline is fill value")
        with exp_warning:
            res = calc.get_nominal_start_time(dt.datetime(2020, 1, 1, 12, 0, 0))
        assert res == expected

    @pytest.mark.parametrize(
        ("area", "expected"),
        [
            (
                "JP01",
                {"tstart": dt.datetime(2018, 10, 22, 3, 0, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 2, 30)}
            ),
            (
                "JP04",
                {"tstart": dt.datetime(2018, 10, 22, 3, 7, 30, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 10, 0, 0)}
            ),
            (
                "R301",
                {"tstart": dt.datetime(2018, 10, 22, 3, 0, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 2, 30)}
            ),
            (
                "R304",
                {"tstart": dt.datetime(2018, 10, 22, 3, 7, 30, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 10, 0, 0)}
            ),
            (
                "R401",
                {"tstart": dt.datetime(2018, 10, 22, 3, 0, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 0, 30)}
            ),
            (
                "R420",
                {"tstart": dt.datetime(2018, 10, 22, 3, 9, 30, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 10, 0, 0)}
            ),
            (
                "R501",
                {"tstart": dt.datetime(2018, 10, 22, 3, 0, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 0, 30)}
            ),
            (
                "R520",
                {"tstart": dt.datetime(2018, 10, 22, 3, 9, 30, 0),
                 "tend": dt.datetime(2018, 10, 22, 3, 10, 0, 0)}
            ),
        ]
    )
    def test_areas(self, area, expected):
        """Test nominal timestamps for multiple areas."""
        obs_start_time = dt.datetime(2018, 10, 22, 3, 0, 20, 596896)
        calc = _NominalTimeCalculator("0300", area)
        nom_start_time = calc.get_nominal_start_time(obs_start_time)
        nom_end_time = calc.get_nominal_end_time(nom_start_time)
        assert nom_start_time == expected["tstart"]
        assert nom_end_time == expected["tend"]

    @pytest.mark.parametrize(
        ("timeline", "obs_start_time", "expected"),
        [
            (
                "2350",
                dt.datetime(2022, 12, 31, 23, 50, 1),
                {"tstart": dt.datetime(2022, 12, 31, 23, 50, 0),
                 "tend": dt.datetime(2023, 1, 1, 0, 0, 0)}
            ),
            (
                "2350",
                dt.datetime(2022, 12, 31, 23, 49, 59),
                {"tstart": dt.datetime(2022, 12, 31, 23, 50, 0),
                 "tend": dt.datetime(2023, 1, 1, 0, 0, 0)}
            ),
            (
                "0000",
                dt.datetime(2023, 1, 1, 0, 0, 1),
                {"tstart": dt.datetime(2023, 1, 1, 0, 0, 0),
                 "tend": dt.datetime(2023, 1, 1, 0, 10, 0)}
            ),
            (
                "0000",
                dt.datetime(2022, 12, 31, 23, 59, 59),
                {"tstart": dt.datetime(2023, 1, 1, 0, 0, 0),
                 "tend": dt.datetime(2023, 1, 1, 0, 10, 0)}
            ),
        ]
    )
    def test_timelines(self, timeline, obs_start_time, expected):
        """Test nominal timestamps for multiple timelines."""
        calc = _NominalTimeCalculator(timeline, "FLDK")
        nom_start_time = calc.get_nominal_start_time(obs_start_time)
        nom_end_time = calc.get_nominal_end_time(nom_start_time)
        assert nom_start_time == expected["tstart"]
        assert nom_end_time == expected["tend"]



"""GMS-5 VISSR Navigation.

Reference: `GMS User Guide`_, Appendix E, S-VISSR Mapping.

.. _GMS User Guide:
    https://www.data.jma.go.jp/mscweb/en/operation/fig/GMS_Users_Guide_3rd_Edition_Rev1.pdf
"""

from collections import namedtuple

import dask.array as da
import numba
import numpy as np

from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()

EARTH_FLATTENING = 1 / 298.257
EARTH_EQUATORIAL_RADIUS = 6378136.0
EARTH_POLAR_RADIUS = EARTH_EQUATORIAL_RADIUS * (1 - EARTH_FLATTENING)
"""Constants taken from JMA's Msial library."""


Pixel = namedtuple(
    "Pixel",
    ["line", "pixel"]
)
"""A VISSR pixel."""

Vector2D = namedtuple(
    "Vector2D",
    ["x", "y"]
)
"""A 2D vector."""


Vector3D = namedtuple(
    "Vector3D",
    ["x", "y", "z"]
)
"""A 3D vector."""


Satpos = namedtuple(
    "Satpos",
    ["x", "y", "z"]
)
"""A 3D vector."""


Attitude = namedtuple(
    "Attitude",
    [
        "angle_between_earth_and_sun",
        "angle_between_sat_spin_and_z_axis",
        "angle_between_sat_spin_and_yz_plane",
    ],
)
"""Attitude parameters.

Units: radians
"""


Orbit = namedtuple(
    "Orbit",
    [
        "angles",
        "sat_position",
        "nutation_precession",
    ],
)
"""Orbital Parameters

Args:
    angles (OrbitAngles): Orbit angles
    sat_position (Vector3D): Satellite position
    nutation_precession: Nutation and precession matrix (3x3)
"""


OrbitAngles = namedtuple(
    "OrbitAngles",
    [
        "greenwich_sidereal_time",
        "declination_from_sat_to_sun",
        "right_ascension_from_sat_to_sun",
    ],
)
"""Orbit angles.

Units: radians
"""


ImageNavigationParameters = namedtuple(
    "ImageNavigationParameters",
    ["static", "predicted"]
)
"""Navigation parameters for the entire image.

Args:
    static (StaticNavigationParameters): Static parameters.
    predicted (PredictedNavigationParameters): Predicted time-dependent parameters.
"""


PixelNavigationParameters = namedtuple(
    "PixelNavigationParameters",
    ["attitude", "orbit", "proj_params"]
)
"""Navigation parameters for a single pixel.

Args:
    attitude (Attitude): Attitude parameters
    orbit (Orbit): Orbit parameters
    proj_params (ProjectionParameters): Projection parameters
"""


StaticNavigationParameters = namedtuple(
    "StaticNavigationParameters",
    [
        "proj_params",
        "scan_params"
    ]
)
"""Navigation parameters which are constant for the entire scan.

Args:
    proj_params (ProjectionParameters): Projection parameters
    scan_params (ScanningParameters): Scanning parameters
"""


PredictedNavigationParameters = namedtuple(
    "PredictedNavigationParameters",
    [
        "attitude",
        "orbit"
    ]
)
"""Predictions of time-dependent navigation parameters.

They need to be evaluated for each pixel.

Args:
    attitude (AttitudePrediction): Attitude prediction
    orbit (OrbitPrediction): Orbit prediction
"""


ScanningParameters = namedtuple(
    "ScanningParameters",
    [
        "start_time_of_scan",
        "spinning_rate",
        "num_sensors",
        "sampling_angle"
    ],
)


ProjectionParameters = namedtuple(
    "ProjectionParameters",
    [
        "image_offset",
        "scanning_angles",
        "earth_ellipsoid",
    ],
)
"""Projection parameters.

Args:
    image_offset (ImageOffset): Image offset
    scanning_angles (ScanningAngles): Scanning angles
    earth_ellipsoid (EarthEllipsoid): Earth ellipsoid
"""


ImageOffset = namedtuple(
    "ImageOffset",
    [
        "line_offset",
        "pixel_offset",
    ]
)
"""Image offset

Args:
    line_offset: Line offset from image center
    pixel_offset: Pixel offset from image center
"""


ScanningAngles = namedtuple(
    "ScanningAngles",
    [
        "stepping_angle",
        "sampling_angle",
        "misalignment"
    ]
)
"""Scanning angles

Args:
    stepping_angle: Scanning angle along line (rad)
    sampling_angle: Scanning angle along pixel (rad)
    misalignment: Misalignment matrix (3x3)
"""


EarthEllipsoid = namedtuple(
    "EarthEllipsoid",
    [
        "flattening",
        "equatorial_radius"
    ]
)
"""Earth ellipsoid.

Args:
    flattening: Ellipsoid flattening
    equatorial_radius: Equatorial radius (meters)
"""


_AttitudePrediction = namedtuple(
    "_AttitudePrediction",
    [
        "prediction_times",
        "attitude"
    ],
)


_OrbitPrediction = namedtuple(
    "_OrbitPrediction",
    [
        "prediction_times",
        "angles",
        "sat_position",
        "nutation_precession",
    ],
)


class AttitudePrediction:
    """Attitude prediction.

    Use .to_numba() to pass this object to jitted methods. This extra
    layer avoids usage of jitclasses and having to re-implement np.unwrap in
    numba.
    """

    def __init__(
        self,
        prediction_times,
        attitude
    ):
        """Initialize attitude prediction.

        In order to accelerate interpolation, the 2-pi periodicity of angles
        is unwrapped here already (that means phase jumps greater than pi
        are wrapped to their 2*pi complement).

        Args:
            prediction_times: Timestamps of predicted attitudes
            attitude (Attitude): Attitudes at prediction times
        """
        self.prediction_times = prediction_times
        self.attitude = self._unwrap_angles(attitude)

    def _unwrap_angles(self, attitude):
        return Attitude(
            np.unwrap(attitude.angle_between_earth_and_sun),
            np.unwrap(attitude.angle_between_sat_spin_and_z_axis),
            np.unwrap(attitude.angle_between_sat_spin_and_yz_plane),
        )

    def to_numba(self):
        """Convert to numba-compatible type."""
        return _AttitudePrediction(
            prediction_times=self.prediction_times,
            attitude=self.attitude
        )


class OrbitPrediction:
    """Orbit prediction.

    Use .to_numba() to pass this object to jitted methods. This extra
    layer avoids usage of jitclasses and having to re-implement np.unwrap in
    numba.
    """

    def __init__(
        self,
        prediction_times,
        angles,
        sat_position,
        nutation_precession,
    ):
        """Initialize orbit prediction.

        In order to accelerate interpolation, the 2-pi periodicity of angles
        is unwrapped here already (that means phase jumps greater than pi
        are wrapped to their 2*pi complement).

        Args:
            prediction_times: Timestamps of orbit prediction.
            angles (OrbitAngles): Orbit angles
            sat_position (Vector3D): Satellite position
            nutation_precession: Nutation and precession matrix.
        """
        self.prediction_times = prediction_times
        self.angles = self._unwrap_angles(angles)
        self.sat_position = sat_position
        self.nutation_precession = nutation_precession

    def _unwrap_angles(self, angles):
        return OrbitAngles(
            greenwich_sidereal_time=np.unwrap(angles.greenwich_sidereal_time),
            declination_from_sat_to_sun=np.unwrap(angles.declination_from_sat_to_sun),
            right_ascension_from_sat_to_sun=np.unwrap(
                angles.right_ascension_from_sat_to_sun
            ),
        )

    def to_numba(self):
        """Convert to numba-compatible type."""
        return _OrbitPrediction(
            prediction_times=self.prediction_times,
            angles=self.angles,
            sat_position=self.sat_position,
            nutation_precession=self.nutation_precession,
        )


def get_lons_lats(lines, pixels, nav_params):
    """Compute lon/lat coordinates given VISSR image coordinates.

    Args:
        lines: VISSR image lines
        pixels: VISSR image pixels
        nav_params: Image navigation parameters
    """
    pixels_2d, lines_2d = da.meshgrid(pixels, lines)
    lons, lats = da.map_blocks(
        _get_lons_lats_numba,
        lines_2d,
        pixels_2d,
        nav_params=_make_nav_params_numba_compatible(nav_params),
        **_get_map_blocks_kwargs(pixels_2d.chunks)
    )
    return lons, lats


def _make_nav_params_numba_compatible(nav_params):
    predicted = PredictedNavigationParameters(
        attitude=nav_params.predicted.attitude.to_numba(),
        orbit=nav_params.predicted.orbit.to_numba()
    )
    return ImageNavigationParameters(nav_params.static, predicted)


def _get_map_blocks_kwargs(chunks):
    # Get keyword arguments for da.map_blocks, so that it can be used
    # with a function that returns two arguments.
    return {
        "new_axis": 0,
        "chunks": (2,) + chunks,
        "dtype": np.float32,
    }


@numba.njit
def _get_lons_lats_numba(lines_2d, pixels_2d, nav_params):
    shape = lines_2d.shape
    lons = np.zeros(shape, dtype=np.float32)
    lats = np.zeros(shape, dtype=np.float32)
    for i in range(shape[0]):
        for j in range(shape[1]):
            pixel = Pixel(lines_2d[i, j], pixels_2d[i, j])
            nav_params_pix = _get_pixel_navigation_parameters(
                pixel, nav_params
            )
            lon, lat = get_lon_lat(pixel, nav_params_pix)
            lons[i, j] = lon
            lats[i, j] = lat
    # Stack lons and lats because da.map_blocks doesn't support multiple
    # return values.
    return np.stack((lons, lats))


@numba.njit
def _get_pixel_navigation_parameters(point, im_nav_params):
    obs_time = get_observation_time(point, im_nav_params.static.scan_params)
    attitude, orbit = interpolate_navigation_prediction(
        attitude_prediction=im_nav_params.predicted.attitude,
        orbit_prediction=im_nav_params.predicted.orbit,
        observation_time=obs_time
    )
    return PixelNavigationParameters(
        attitude=attitude,
        orbit=orbit,
        proj_params=im_nav_params.static.proj_params
    )


@numba.njit
def get_observation_time(point, scan_params):
    """Calculate observation time of a VISSR pixel."""
    relative_time = _get_relative_observation_time(point, scan_params)
    return scan_params.start_time_of_scan + relative_time


@numba.njit
def _get_relative_observation_time(point, scan_params):
    line, pixel = point
    pixel = pixel + 1
    line = line + 1
    spinning_freq = 1440 * scan_params.spinning_rate
    line_step = np.floor((line - 1) / scan_params.num_sensors)
    pixel_step = (scan_params.sampling_angle * pixel) / (2 * np.pi)
    return (line_step + pixel_step) / spinning_freq


@numba.njit
def interpolate_navigation_prediction(
    attitude_prediction, orbit_prediction, observation_time
):
    """Interpolate predicted navigation parameters."""
    attitude = interpolate_attitude_prediction(attitude_prediction, observation_time)
    orbit = interpolate_orbit_prediction(orbit_prediction, observation_time)
    return attitude, orbit


@numba.njit
def get_lon_lat(pixel, nav_params):
    """Get longitude and latitude coordinates for a given image pixel.

    Args:
        pixel (Pixel): Point in image coordinates.
        nav_params (PixelNavigationParameters): Navigation parameters for a
            single pixel.

    Returns:
        Longitude and latitude in degrees.
    """
    scan_angles = transform_image_coords_to_scanning_angles(
        pixel,
        nav_params.proj_params.image_offset,
        nav_params.proj_params.scanning_angles
    )
    view_vector_sat = transform_scanning_angles_to_satellite_coords(
        scan_angles,
        nav_params.proj_params.scanning_angles.misalignment
    )
    view_vector_earth_fixed = transform_satellite_to_earth_fixed_coords(
        view_vector_sat,
        nav_params.orbit,
        nav_params.attitude
    )
    point_on_earth = intersect_with_earth(
        view_vector_earth_fixed,
        nav_params.orbit.sat_position,
        nav_params.proj_params.earth_ellipsoid
    )
    lon, lat = transform_earth_fixed_to_geodetic_coords(
        point_on_earth, nav_params.proj_params.earth_ellipsoid.flattening
    )
    return lon, lat


@numba.njit
def transform_image_coords_to_scanning_angles(point, image_offset, scanning_angles):
    """Transform image coordinates to scanning angles.

    Args:
        point (Pixel): Point in image coordinates.
        image_offset (ImageOffset): Image offset.
        scanning_angles (ScanningAngles): Scanning angles.

    Returns:
        Scanning angles (x, y) at the pixel center (rad).
    """
    line_offset = image_offset.line_offset
    pixel_offset = image_offset.pixel_offset
    stepping_angle = scanning_angles.stepping_angle
    sampling_angle = scanning_angles.sampling_angle
    x = sampling_angle * (point.pixel + 1 - pixel_offset)
    y = stepping_angle * (point.line + 1 - line_offset)
    return Vector2D(x, y)


@numba.njit
def transform_scanning_angles_to_satellite_coords(angles, misalignment):
    """Transform scanning angles to satellite angular momentum coordinates.

    Args:
        angles (Vector2D): Scanning angles in radians.
        misalignment: Misalignment matrix (3x3)

    Returns:
        View vector (Vector3D) in satellite angular momentum coordinates.
    """
    x, y = angles.x, angles.y
    sin_x = np.sin(x)
    cos_x = np.cos(x)
    view = Vector3D(np.cos(y), 0.0, np.sin(y))

    # Correct for misalignment
    view = matrix_vector(misalignment, view)

    # Rotate around z-axis
    return Vector3D(
        cos_x * view.x - sin_x * view.y,
        sin_x * view.x + cos_x * view.y,
        view.z
    )


@numba.njit
def transform_satellite_to_earth_fixed_coords(
    point,
    orbit,
    attitude
):
    """Transform from earth-fixed to satellite angular momentum coordinates.

    Args:
        point (Vector3D): Point in satellite angular momentum coordinates.
        orbit (Orbit): Orbital parameters
        attitude (Attitude): Attitude parameters
    Returns:
        Point (Vector3D) in earth-fixed coordinates.
    """
    unit_vector_z = _get_satellite_unit_vector_z(attitude, orbit)
    unit_vector_x = _get_satellite_unit_vector_x(unit_vector_z, attitude, orbit)
    unit_vector_y = _get_satellite_unit_vector_y(unit_vector_x, unit_vector_z)
    return _get_earth_fixed_coords(
        point,
        unit_vector_x,
        unit_vector_y,
        unit_vector_z
    )


@numba.njit
def _get_satellite_unit_vector_z(attitude, orbit):
    v1950 = _get_satellite_z_axis_1950(
        attitude.angle_between_sat_spin_and_z_axis,
        attitude.angle_between_sat_spin_and_yz_plane
    )
    vcorr = _correct_nutation_precession(
        v1950,
        orbit.nutation_precession
    )
    return _rotate_to_greenwich(
        vcorr,
        orbit.angles.greenwich_sidereal_time
    )


@numba.njit
def _get_satellite_z_axis_1950(
    angle_between_sat_spin_and_z_axis,
    angle_between_sat_spin_and_yz_plane
):
    """Get satellite z-axis (spin) in mean of 1950 coordinates."""
    alpha = angle_between_sat_spin_and_z_axis
    delta = angle_between_sat_spin_and_yz_plane
    cos_delta = np.cos(delta)
    return Vector3D(
        x=np.sin(delta),
        y=-cos_delta * np.sin(alpha),
        z=cos_delta * np.cos(alpha)
    )


@numba.njit
def _correct_nutation_precession(vector, nutation_precession):
    return matrix_vector(nutation_precession, vector)


@numba.njit
def _rotate_to_greenwich(vector, greenwich_sidereal_time):
    cos_sid = np.cos(greenwich_sidereal_time)
    sin_sid = np.sin(greenwich_sidereal_time)
    rotated = Vector3D(
        x=cos_sid * vector.x + sin_sid * vector.y,
        y=-sin_sid * vector.x + cos_sid * vector.y,
        z=vector.z
    )
    return normalize_vector(rotated)


@numba.njit
def _get_satellite_unit_vector_x(unit_vector_z, attitude, orbit):
    sat_sun_vec = _get_vector_from_satellite_to_sun(
        orbit.angles.declination_from_sat_to_sun,
        orbit.angles.right_ascension_from_sat_to_sun
    )
    return _get_unit_vector_x(
        sat_sun_vec,
        unit_vector_z,
        attitude.angle_between_earth_and_sun
    )


@numba.njit
def _get_vector_from_satellite_to_sun(
    declination_from_sat_to_sun,
    right_ascension_from_sat_to_sun
):
    declination = declination_from_sat_to_sun
    right_ascension = right_ascension_from_sat_to_sun
    cos_declination = np.cos(declination)
    return Vector3D(
        x=cos_declination * np.cos(right_ascension),
        y=cos_declination * np.sin(right_ascension),
        z=np.sin(declination)
    )


@numba.njit
def _get_unit_vector_x(
    sat_sun_vec,
    unit_vector_z,
    angle_between_earth_and_sun

):
    beta = angle_between_earth_and_sun
    sin_beta = np.sin(beta)
    cos_beta = np.cos(beta)
    cross1 = _get_uz_cross_satsun(unit_vector_z, sat_sun_vec)
    cross2 = cross_product(cross1, unit_vector_z)
    unit_vector_x = Vector3D(
        x=sin_beta * cross1.x + cos_beta * cross2.x,
        y=sin_beta * cross1.y + cos_beta * cross2.y,
        z=sin_beta * cross1.z + cos_beta * cross2.z
    )
    return normalize_vector(unit_vector_x)


@numba.njit
def _get_uz_cross_satsun(unit_vector_z, sat_sun_vec):
    res = cross_product(unit_vector_z, sat_sun_vec)
    return normalize_vector(res)


@numba.njit
def _get_satellite_unit_vector_y(unit_vector_x, unit_vector_z):
    res = cross_product(unit_vector_z, unit_vector_x)
    return normalize_vector(res)


@numba.njit
def _get_earth_fixed_coords(point, unit_vector_x, unit_vector_y, unit_vector_z):
    ux, uy, uz = unit_vector_x, unit_vector_y, unit_vector_z
    # Multiply with matrix of satellite unit vectors [ux, uy, uz]
    return Vector3D(
        x=ux.x * point.x + uy.x * point.y + uz.x * point.z,
        y=ux.y * point.x + uy.y * point.y + uz.y * point.z,
        z=ux.z * point.x + uy.z * point.y + uz.z * point.z
    )


@numba.njit
def intersect_with_earth(view_vector, sat_pos, ellipsoid):
    """Intersect instrument viewing vector with the earth's surface.

    Reference: Appendix E, section 2.11 in the GMS user guide.

    Args:
        view_vector (Vector3D): Instrument viewing vector in earth-fixed
            coordinates.
        sat_pos (Vector3D): Satellite position in earth-fixed coordinates.
        ellipsoid (EarthEllipsoid): Earth ellipsoid.

    Returns:
        Intersection (Vector3D) with the earth's surface.
    """
    distance = _get_distance_to_intersection(view_vector, sat_pos, ellipsoid)
    return Vector3D(
        sat_pos.x + distance * view_vector.x,
        sat_pos.y + distance * view_vector.y,
        sat_pos.z + distance * view_vector.z
    )


@numba.njit
def _get_distance_to_intersection(view_vector, sat_pos, ellipsoid):
    """Get distance to intersection with the earth.

    If the instrument is pointing towards the earth, there will be two
    intersections with the surface. Choose the one on the instrument-facing
    side of the earth.
    """
    d1, d2 = _get_distances_to_intersections(view_vector, sat_pos, ellipsoid)
    return min(d1, d2)


@numba.njit
def _get_distances_to_intersections(view_vector, sat_pos, ellipsoid):
    """Get distances to intersections with the earth's surface.

    Returns:
        Distances to two intersections with the surface.
    """
    a, b, c = _get_abc_helper(view_vector, sat_pos, ellipsoid)
    tmp = np.sqrt((b**2 - a * c))
    dist_1 = (-b + tmp) / a
    dist_2 = (-b - tmp) / a
    return dist_1, dist_2


@numba.njit
def _get_abc_helper(view_vector, sat_pos, ellipsoid):
    """Get a,b,c helper variables.

    Reference: Appendix E, Equation (26) in the GMS user guide.
    """
    flat2 = (1 - ellipsoid.flattening) ** 2
    ux, uy, uz = view_vector.x, view_vector.y, view_vector.z
    x, y, z = sat_pos.x, sat_pos.y, sat_pos.z
    a = flat2 * (ux ** 2 + uy ** 2) + uz ** 2
    b = flat2 * (x * ux + y * uy) + z * uz
    c = flat2 * (x ** 2 + y ** 2 - ellipsoid.equatorial_radius ** 2) + z ** 2
    return a, b, c


@numba.njit
def transform_earth_fixed_to_geodetic_coords(point, earth_flattening):
    """Transform from earth-fixed to geodetic coordinates.

    Args:
        point (Vector3D): Point in earth-fixed coordinates.
        earth_flattening: Flattening of the earth.

    Returns:
        Geodetic longitude and latitude (degrees).
    """
    x, y, z = point.x, point.y, point.z
    f = earth_flattening
    lon = np.arctan2(y, x)
    lat = np.arctan2(z, ((1 - f) ** 2 * np.sqrt(x**2 + y**2)))
    return np.rad2deg(lon), np.rad2deg(lat)


@numba.njit
def interpolate_orbit_prediction(orbit_prediction, observation_time):
    """Interpolate orbit prediction at the given observation time."""
    angles = _interpolate_orbit_angles(observation_time, orbit_prediction)
    sat_position = _interpolate_sat_position(observation_time, orbit_prediction)
    nutation_precession = interpolate_nearest(
        observation_time,
        orbit_prediction.prediction_times,
        orbit_prediction.nutation_precession,
    )
    return Orbit(
        angles=angles,
        sat_position=sat_position,
        nutation_precession=nutation_precession,
    )


@numba.njit
def _interpolate_orbit_angles(observation_time, orbit_prediction):
    sidereal_time = interpolate_angles(
        observation_time,
        orbit_prediction.prediction_times,
        orbit_prediction.angles.greenwich_sidereal_time,
    )
    declination = interpolate_angles(
        observation_time,
        orbit_prediction.prediction_times,
        orbit_prediction.angles.declination_from_sat_to_sun,
    )
    right_ascension = interpolate_angles(
        observation_time,
        orbit_prediction.prediction_times,
        orbit_prediction.angles.right_ascension_from_sat_to_sun,
    )
    return OrbitAngles(
        greenwich_sidereal_time=sidereal_time,
        declination_from_sat_to_sun=declination,
        right_ascension_from_sat_to_sun=right_ascension,
    )


@numba.njit
def _interpolate_sat_position(observation_time, orbit_prediction):
    x = interpolate_continuous(
        observation_time,
        orbit_prediction.prediction_times,
        orbit_prediction.sat_position.x,
    )
    y = interpolate_continuous(
        observation_time,
        orbit_prediction.prediction_times,
        orbit_prediction.sat_position.y,
    )
    z = interpolate_continuous(
        observation_time,
        orbit_prediction.prediction_times,
        orbit_prediction.sat_position.z,
    )
    return Vector3D(x, y, z)


@numba.njit
def interpolate_attitude_prediction(attitude_prediction, observation_time):
    """Interpolate attitude prediction at given observation time."""
    angle_between_earth_and_sun = interpolate_angles(
        observation_time,
        attitude_prediction.prediction_times,
        attitude_prediction.attitude.angle_between_earth_and_sun,
    )
    angle_between_sat_spin_and_z_axis = interpolate_angles(
        observation_time,
        attitude_prediction.prediction_times,
        attitude_prediction.attitude.angle_between_sat_spin_and_z_axis,
    )
    angle_between_sat_spin_and_yz_plane = interpolate_angles(
        observation_time,
        attitude_prediction.prediction_times,
        attitude_prediction.attitude.angle_between_sat_spin_and_yz_plane,
    )
    return Attitude(
        angle_between_earth_and_sun,
        angle_between_sat_spin_and_z_axis,
        angle_between_sat_spin_and_yz_plane,
    )


@numba.njit
def interpolate_continuous(x, x_sample, y_sample):
    """Linear interpolation of continuous quantities.

    Numpy equivalent would be np.interp(..., left=np.nan, right=np.nan), but
    numba currently doesn't support those keyword arguments.
    """
    try:
        return _interpolate(x, x_sample, y_sample)
    except Exception:
        # Numba cannot distinguish exception types
        return np.nan


@numba.njit
def _interpolate(x, x_sample, y_sample):
    i = _find_enclosing_index(x, x_sample)
    offset = y_sample[i]
    x_diff = x_sample[i + 1] - x_sample[i]
    y_diff = y_sample[i + 1] - y_sample[i]
    slope = y_diff / x_diff
    dist = x - x_sample[i]
    return offset + slope * dist


@numba.njit
def _find_enclosing_index(x, x_sample):
    """Find where x_sample encloses x."""
    for i in range(len(x_sample) - 1):
        if x_sample[i] <= x < x_sample[i + 1]:
            return i
    raise Exception("x not enclosed by x_sample")


@numba.njit
def interpolate_angles(x, x_sample, y_sample):
    """Linear interpolation of angles.

    Requires 2-pi periodicity to be unwrapped before (for
    performance reasons). Interpolated angles are wrapped
    back to [-pi, pi] to restore periodicity.
    """
    return _wrap_2pi(interpolate_continuous(x, x_sample, y_sample))


@numba.njit
def _wrap_2pi(values):
    """Wrap values to interval [-pi, pi].

    Source: https://stackoverflow.com/a/15927914/5703449
    """
    return (values + np.pi) % (2 * np.pi) - np.pi


@numba.njit
def interpolate_nearest(x, x_sample, y_sample):
    """Nearest neighbour interpolation."""
    try:
        return _interpolate_nearest(x, x_sample, y_sample)
    except Exception:
        return np.nan * np.ones_like(y_sample[0])


@numba.njit
def _interpolate_nearest(x, x_sample, y_sample):
    i = _find_enclosing_index(x, x_sample)
    return y_sample[i]


@numba.njit
def matrix_vector(m, v):
    """Multiply (3,3)-matrix and Vector3D."""
    x = m[0, 0] * v.x + m[0, 1] * v.y + m[0, 2] * v.z
    y = m[1, 0] * v.x + m[1, 1] * v.y + m[1, 2] * v.z
    z = m[2, 0] * v.x + m[2, 1] * v.y + m[2, 2] * v.z
    return Vector3D(x, y, z)


@numba.njit
def cross_product(a, b):
    """Compute vector product a x b."""
    return Vector3D(
        x=a.y * b.z - a.z * b.y,
        y=a.z * b.x - a.x * b.z,
        z=a.x * b.y - a.y * b.x
    )


@numba.njit
def normalize_vector(v):
    """Normalize a Vector3D."""
    norm = np.sqrt(v.x**2 + v.y**2 + v.z**2)
    return Vector3D(
        v.x / norm,
        v.y / norm,
        v.z / norm
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Pygac interface."""
import datetime as dt
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import numpy as np
import pygac.gac_klm
import pygac.gac_pod
import pygac.klm_reader
import pygac.pod_reader
import pytest
import xarray as xr
from pygac.gac_reader import GACReader

from satpy import Scene
from satpy.readers.avhrr_l1b_gaclac import GACLACFile

GAC_PATTERN = '{creation_site:3s}.{transfer_mode:4s}.{platform_id:2s}.D{start_time:%y%j.S%H%M}.E{end_time:%H%M}.B{orbit_number:05d}{end_orbit_last_digits:02d}.{station:2s}'  # noqa
EOSIP_PATTERN = '{platform_id:3s}_RPRO_AVH_L1B_1P_{start_time:%Y%m%dT%H%M%S}_{end_time:%Y%m%dT%H%M%S}_{orbit_number:06d}/image.l1b'  # noqa

GAC_POD_FILENAMES = ["NSS.GHRR.NA.D79184.S1150.E1337.B0008384.WI",
                     "NSS.GHRR.NA.D79184.S2350.E0137.B0008384.WI",
                     "NSS.GHRR.NA.D80021.S0927.E1121.B0295354.WI",
                     "NSS.GHRR.NA.D80021.S1120.E1301.B0295455.WI",
                     "NSS.GHRR.NA.D80021.S1256.E1450.B0295556.GC",
                     "NSS.GHRR.NE.D83208.S1219.E1404.B0171819.WI",
                     "NSS.GHRR.NG.D88002.S0614.E0807.B0670506.WI",
                     "NSS.GHRR.TN.D79183.S1258.E1444.B0369697.GC",
                     "NSS.GHRR.TN.D80003.S1147.E1332.B0630506.GC",
                     "NSS.GHRR.TN.D80003.S1328.E1513.B0630507.GC",
                     "NSS.GHRR.TN.D80003.S1509.E1654.B0630608.GC"]

GAC_KLM_FILENAMES = ["NSS.GHRR.NK.D01235.S0252.E0446.B1703233.GC",
                     "NSS.GHRR.NL.D01288.S2315.E0104.B0549495.GC",
                     "NSS.GHRR.NM.D04111.S2305.E0050.B0947778.GC",
                     "NSS.GHRR.NN.D13011.S0559.E0741.B3939192.WI",
                     "NSS.GHRR.NP.D15361.S0121.E0315.B3547172.SV",
                     "NSS.GHRR.M1.D15362.S0031.E0129.B1699697.SV",
                     "NSS.GHRR.M2.D10178.S2359.E0142.B1914142.SV"]

LAC_POD_FILENAMES = ["BRN.HRPT.ND.D95152.S1730.E1715.B2102323.UB",
                     "BRN.HRPT.ND.D95152.S1910.E1857.B2102424.UB",
                     "BRN.HRPT.NF.D85152.S1345.E1330.B0241414.UB",
                     "BRN.HRPT.NJ.D95152.S1233.E1217.B0216060.UB"]

LAC_KLM_FILENAMES = ["BRN.HRPT.M1.D14152.S0958.E1012.B0883232.UB",
                     "BRN.HRPT.M1.D14152.S1943.E1958.B0883838.UB",
                     "BRN.HRPT.M2.D12153.S0912.E0922.B2914747.UB",
                     "BRN.HRPT.NN.D12153.S0138.E0152.B3622828.UB",
                     "BRN.HRPT.NN.D12153.S0139.E0153.B3622828.UB",
                     "BRN.HRPT.NN.D12153.S1309.E1324.B3623535.UB",
                     "BRN.HRPT.NP.D12153.S0003.E0016.B1707272.UB",
                     "BRN.HRPT.NP.D12153.S1134.E1148.B1707979.UB",
                     "BRN.HRPT.NP.D16184.S1256.E1311.B3813131.UB",
                     "BRN.HRPT.NP.D16184.S1438.E1451.B3813232.UB",
                     "BRN.HRPT.NP.D16184.S1439.E1451.B3813232.UB",
                     "BRN.HRPT.NP.D16185.S1245.E1259.B3814545.UB",
                     "BRN.HRPT.NP.D16185.S1427.E1440.B3814646.UB",
                     "NSS.FRAC.M2.D12153.S1729.E1910.B2915354.SV",
                     "NSS.LHRR.NP.D16306.S1803.E1814.B3985555.WI"]

LAC_EOSIP_FILENAMES = ["N06_RPRO_AVH_L1B_1P_20061206T010808_20061206T012223_007961/image.l1b"]


class TestGACLACFile:
    """Test the GACLAC file handler."""

    def _get_fh(self, filename="NSS.GHRR.NG.D88002.S0614.E0807.B0670506.WI",
                **kwargs):
        """Create a file handler."""
        from trollsift import parse
        filename_info = parse(GAC_PATTERN, filename)
        return GACLACFile(filename, filename_info, {}, **kwargs)

    def _get_eosip_fh(self, filename, **kwargs):
        """Create a file handler."""
        from trollsift import parse
        filename_info = parse(EOSIP_PATTERN, filename)
        return GACLACFile(filename, filename_info, {}, **kwargs)

    def test_init(self):
        """Test GACLACFile initialization."""
        from pygac.gac_klm import GACKLMReader
        from pygac.gac_pod import GACPODReader
        from pygac.lac_klm import LACKLMReader
        from pygac.lac_pod import LACPODReader

        kwargs = {"start_line": 1,
                  "end_line": 2,
                  "strip_invalid_coords": True,
                  "interpolate_coords": True,
                  "adjust_clock_drift": True,
                  "tle_dir": "tle_dir",
                  "tle_name": "tle_name",
                  "tle_thresh": 123,
                  "calibration": "calibration"}
        for filenames, reader_cls in zip([GAC_POD_FILENAMES, GAC_KLM_FILENAMES, LAC_POD_FILENAMES, LAC_KLM_FILENAMES],
                                         [GACPODReader, GACKLMReader, LACPODReader, LACKLMReader]):
            for filename in filenames:
                fh = self._get_fh(filename, **kwargs)
                assert fh.start_time < fh.end_time
                assert fh.reader_class is reader_cls


    def test_init_eosip(self):
        """Test GACLACFile initialization."""
        from pygac.lac_pod import LACPODReader

        kwargs = {"start_line": 1,
                  "end_line": 2,
                  "strip_invalid_coords": True,
                  "interpolate_coords": True,
                  "adjust_clock_drift": True,
                  "tle_dir": "tle_dir",
                  "tle_name": "tle_name",
                  "tle_thresh": 123,
                  "calibration": "calibration"}
        for filenames, reader_cls in zip([LAC_EOSIP_FILENAMES],
                                         [LACPODReader]):
            for filename in filenames:
                fh = self._get_eosip_fh(filename, **kwargs)
                assert fh.start_time < fh.end_time
                assert fh.reader_class is reader_cls
                assert fh.reader_kwargs["header_date"] > dt.date(1994, 11, 15)



class DataType(Enum):
    """AVHRR GAC data type."""
    KLM = "klm"
    POD = "pod"


@dataclass
class TestParams:
    """Test parameters."""
    data_type: DataType
    satellite: str
    tle: str
    filename: str
    reader_kwargs: dict
    qual_flag: int = 0

@dataclass
class Expectations:
    """Expected scene properties."""
    num_lines: int
    start_time: dt.datetime
    end_time: dt.datetime
    num_lines_slc: int
    start_time_slc: dt.datetime
    end_time_slc: dt.datetime
    platform_name: str


tle_noaa14 = """1 23455U 94089A   09363.50221683 -.00000013  00000-0  16673-4 0  8853
2 23455  98.8939  80.3339 0009410   2.6550 357.4665 14.13775739773522
1 23455U 94089A   09363.50221683 -.00000013  00000-0  16673-4 0  8864
2 23455  98.8939  80.3339 0009410   2.6550 357.4665 14.13775739773522
1 23455U 94089A   09363.78530466 -.00000027  00000-0  94367-5 0  8879
2 23455  98.8938  80.6155 0009463   1.7605 358.3582 14.13775588773563"""

tle_noaa15 = """1 25338U 98030A   09361.80631861 -.00000112  00000-0 -29536-4 0  2104
2 25338  98.6031 346.9543 0010160 209.9249 150.1326 14.24798410604342
1 25338U 98030A   09362.36812158 -.00000012  00000-0  13243-4 0  2102
2 25338  98.6030 347.5048 0010160 208.4397 151.6285 14.24799183604423
1 25338U 98030A   09363.49172435  .00000077  00000-0  51611-4 0  2116
2 25338  98.6031 348.6054 0010232 205.1043 154.9644 14.24799790604581"""



class FakeDataGenerator:
    """Generate fake GAC data."""

    @staticmethod
    def get_data(params: TestParams) -> list[np.ndarray]:
        """Get fake data."""
        methods = {
            DataType.KLM: FakeDataGenerator._get_klm_data,
            DataType.POD: FakeDataGenerator._get_pod_data,
        }
        return methods[params.data_type](params.filename, params.qual_flag)

    @staticmethod
    def _get_klm_data(filename: str, qual_flag: int):
        num_lines = 55
        times, msecs_since_00 = FakeDataGenerator._get_times(num_lines)
        telemetry = np.zeros(1, dtype=pygac.klm_reader.analog_telemetry_v2)
        scans = np.ones(num_lines, dtype=pygac.gac_klm.scanline)
        scans["scan_line_number"] = np.arange(num_lines)
        scans["scan_line_year"] = times.dt.year.values
        scans["scan_line_day_of_year"] = times.dt.dayofyear.values
        scans["scan_line_utc_time_of_day"] = msecs_since_00.values
        scans["telemetry"]["PRT"] = np.repeat(np.arange(num_lines), 3).reshape((num_lines, 3))
        scans["quality_indicator_bit_field"] = np.full(num_lines, qual_flag)

        hdr = np.zeros(1, dtype=pygac.klm_reader.header)
        hdr["data_type_code"] = 2  # GAC
        hdr["data_set_name"] = filename
        hdr["noaa_spacecraft_identification_code"] = 4  # NOAA-15
        hdr["noaa_level_1b_format_version_number"] = 2
        hdr["count_of_data_records"] = num_lines
        hdr["start_of_data_set_year"] = times.dt.year.values[0]
        hdr["start_of_data_set_day_of_year"] = times.dt.dayofyear.values[0]
        hdr["start_of_data_set_utc_time_of_day"] = msecs_since_00.values[0]

        spare = np.zeros(pygac.gac_klm.GACKLMReader().offset - (hdr.itemsize + telemetry.itemsize), dtype="u1")

        return [hdr, telemetry, spare, scans]

    @staticmethod
    def _get_times(num_lines: int):
        times = xr.DataArray(
            [
                dt.datetime(2009, 12, 28, 23, 59, 50) + dt.timedelta(milliseconds=line/GACReader.scan_freq)
                for line in range(num_lines)
            ],
        )
        msecs_since_00 = (
            1000 * (times.dt.hour * 3600 + times.dt.minute * 60 + times.dt.second) +
            times.dt.microsecond / 1000
        )
        return times, msecs_since_00

    @staticmethod
    def _get_pod_data(filename: str, qual_flag: int):
        num_lines = 55
        times, msecs_since_00 = FakeDataGenerator._get_times(num_lines)
        times_enc = encode_timestamps_pod(
            times.dt.year.values,
            times.dt.dayofyear.values,
            msecs_since_00.values.astype(int)
        )

        scans = np.ones(num_lines, dtype=pygac.gac_pod.scanline)
        scans["scan_line_number"] = np.arange(num_lines)
        scans["time_code"] = times_enc
        scans["telemetry"] = 100 * np.arange(35 * num_lines).reshape((num_lines, 35))
        scans["quality_indicators"] = np.empty(num_lines, np.uint16(qual_flag))

        hdr0 = np.zeros(1, dtype=pygac.pod_reader.header0)
        hdr0["start_time"] = times_enc[0]
        hdr0["number_of_scans"] = num_lines
        hdr0["noaa_spacecraft_identification_code"] = 3  # NOAA-14

        hdr3 = np.zeros(1, dtype=pygac.pod_reader.header3)

        spare = np.zeros(pygac.gac_pod.GACPODReader().offset - (hdr0.itemsize + hdr3.itemsize), dtype="u1")
        return [hdr0, hdr3, spare, scans]


def encode_timestamps_pod(year: np.ndarray, jday: np.ndarray, msec: np.ndarray) -> np.ndarray:
    """Encode timestamps like in POD files.

    Reverse engineered from Pygac code by an LLM.
    """
    yoff = year - 2000  # assuming we're in 2009 for this test case
    enc0 = (yoff << 9) | (jday & 0x1FF)
    enc2 = msec & 0xFFFF
    enc1 = (msec >> 16) & 0x7FF
    return np.array([enc0, enc1, enc2]).transpose()


def test_encode_timestamps_pod():
    """Test timestamp encoding/decoding round trip."""
    year = np.array([2009, 2009])
    jday = np.array([362, 362])
    msec = np.array([0, 123456])
    encoded = encode_timestamps_pod(year, jday, msec)
    year_dec, jday_dec, msec_dec = pygac.pod_reader.PODReader.decode_timestamps(encoded)
    np.testing.assert_equal(year_dec, year)
    np.testing.assert_equal(jday_dec, jday)
    np.testing.assert_equal(msec_dec, msec)


@pytest.fixture(scope="class")
def tmp_dir(tmp_path_factory):
    """Create temporary directory."""
    return tmp_path_factory.mktemp("stubs")


@pytest.fixture(scope="class")
def tle_dir(tmp_dir: Path):
    """Create TLE directory."""
    tle_dir = tmp_dir / "tle"
    tle_dir.mkdir()
    return tle_dir



def _write_tle(params, tle_dir):
    filename = tle_dir / f"TLE_{params.satellite}.txt"
    with filename.open("w") as fh:
        fh.write(params.tle)


def _write_stub(params: TestParams, tmp_dir: Path):
    fake_data = FakeDataGenerator.get_data(params)
    filename = tmp_dir / params.filename
    with filename.open("wb") as fh:
        for array in fake_data:
            array.tofile(fh)
    return filename


def _get_reader_kwargs(params: TestParams, tle_dir: Path):
    default = {"tle_name": "TLE_%(satname)s.txt", "tle_dir": str(tle_dir)}
    return default | params.reader_kwargs


class TestReadingGacFile:
    """Test reading GAC files."""

    klm_params = TestParams(
        data_type=DataType.KLM,
        satellite="noaa15",
        filename="NSS.GHRR.NK.D09362.S2359.E0001.B6044445.GC",
        reader_kwargs={"strip_invalid_coords": True},
        tle=tle_noaa15
    )
    pod_params = TestParams(
        data_type=DataType.POD,
        satellite="noaa14",
        filename="NSS.GHRR.NJ.D09362.S2359.E0001.B6044445.GC",
        # When stripping invaid coordinates the remaining test sample is too
        # small for the POD reader, so turn it off.
        reader_kwargs={"strip_invalid_coords": False},
        tle=tle_noaa14
    )

    @pytest.fixture(params=["klm_params", "pod_params"], scope="class")
    def params(self, request):
        """Get test parameters."""
        return getattr(request.cls, request.param)

    @pytest.fixture(scope="class")
    def expect(self, params: TestParams):
        """Get expectations."""
        klm_expect = Expectations(
            num_lines=55,
            start_time=dt.datetime(2009, 12, 29, 9, 5, 57, 500000),
            end_time=dt.datetime(2009, 12, 29, 0, 0, 16, 500000),
            num_lines_slc=11,
            start_time_slc=dt.datetime(2009, 12, 28, 23, 59, 54, 500000),
            end_time_slc=dt.datetime(2009, 12, 28, 23, 59, 59, 500000),
            platform_name="noaa15"
        )
        pod_expect = Expectations(
            num_lines=54,
            start_time=dt.datetime(2009, 12, 28, 23, 59, 49, 800000),
            end_time=dt.datetime(2009, 12, 29, 0, 0, 16, 300000),
            num_lines_slc=11,
            start_time_slc=dt.datetime(2009, 12, 28, 23, 59, 54, 800000),
            end_time_slc=dt.datetime(2009, 12, 28, 23, 59, 59, 800000),
            platform_name="noaa14"
        )
        exp = {
            DataType.KLM: klm_expect,
            DataType.POD: pod_expect
        }
        return exp[params.data_type]

    @pytest.fixture(autouse=True, scope="class")
    def tle_file(self, tle_dir: Path, params: TestParams):
        """Write TLE file."""
        _write_tle(params, tle_dir)

    @pytest.fixture(scope="class")
    def stub(self, tmp_dir: Path, params: TestParams):
        """Write stub file."""
        return _write_stub(params, tmp_dir)

    @pytest.fixture(scope="class")
    def reader_kwargs(self, params: TestParams, tle_dir: Path):
        """Get reader keyword arguments."""
        return _get_reader_kwargs(params, tle_dir)

    def test_get_channel_calibrated(self, stub: Path, reader_kwargs: dict, expect: Expectations):
        """Test getting calibrated channel."""
        scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                      reader_kwargs=reader_kwargs)
        scene.load(["1"])
        assert scene["1"].shape == (expect.num_lines, 409)
        assert scene["1"].dims == ("y", "x")
        assert scene["1"].start_time == expect.start_time
        assert scene["1"].end_time == expect.end_time
        assert scene["1"].units == "%"
        assert "tle" in scene["1"].attrs["orbital_parameters"]
        assert scene["1"].attrs["platform_name"] == expect.platform_name

    def test_get_channel_counts(self, stub: Path, reader_kwargs: dict):
        """Test getting raw channel counts."""
        scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                      reader_kwargs=reader_kwargs)
        scene.load(["1"], calibration="counts")
        assert scene["1"].units == "count"

    def test_slice(self, stub: Path, reader_kwargs: dict, expect: Expectations):
        """Test slicing a range of scanlines from the overpass."""
        slice_kwargs = reader_kwargs | {"start_line": 10, "end_line": 20}
        scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                      reader_kwargs=slice_kwargs)
        scene.load(["1"])
        assert scene["1"].shape == (expect.num_lines_slc, 409)
        assert scene["1"].start_time == expect.start_time_slc
        assert scene["1"].end_time == expect.end_time_slc

    def test_get_latlon(self, stub: Path, reader_kwargs: dict, expect: Expectations):
        """Test getting lat/lon coordinates."""
        scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                      reader_kwargs=reader_kwargs)
        scene.load(["latitude"])
        assert scene["latitude"].shape == (expect.num_lines, 409)
        assert scene["latitude"].dims == ("y", "x")
        assert scene["latitude"].units == "degrees_north"

    def test_get_angles(self, stub: Path, reader_kwargs: dict, expect: Expectations):
        """Test getting sun/sensor angles."""
        scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                      reader_kwargs=reader_kwargs)
        scene.load(["sensor_zenith_angle"])
        assert scene["sensor_zenith_angle"].shape == (expect.num_lines, 409)
        assert scene["sensor_zenith_angle"].dims == ("y", "x")
        assert scene["sensor_zenith_angle"].units == "degrees"

    def test_get_qual_flags(self, stub: Path, reader_kwargs: dict, expect: Expectations):
        """Test getting quality flags."""
        scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                      reader_kwargs=reader_kwargs)
        scene.load(["qual_flags"])
        assert scene["qual_flags"].shape == (expect.num_lines, 7)
        assert scene["qual_flags"].dims == ("y", "num_flags")

    def test_get_latlon_without_interp(self, stub: Path, params: TestParams, expect: Expectations, tle_dir: Path):
        """Test getting lat/lon coordinates without interpolation.

        Only works for KLM, because the stub file is too small.
        """
        if params.data_type == DataType.POD:
            pytest.skip()

        params.reader_kwargs = {"interpolate_coords": False}
        reader_kwargs = _get_reader_kwargs(params, tle_dir)
        scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                      reader_kwargs=reader_kwargs)
        scene.load(["latitude"])
        assert scene["latitude"].shape == (expect.num_lines, 51)
        assert scene["latitude"].dims == ("y", "x_every_eighth")


def test_all_data_masked_out(tmp_dir: Path, tle_dir: Path):
    """Test reading a file where all scanlines are masked."""
    flags = pygac.klm_reader.KLM_QualityIndicator
    params = TestParams(
        data_type=DataType.KLM,
        satellite="noaa15",
        tle=tle_noaa15,
        filename="NSS.GHRR.NK.D09362.S2359.E0001.B6044445.GC",
        reader_kwargs={},
        qual_flag=flags.FATAL_FLAG | flags.CALIBRATION | flags.NO_EARTH_LOCATION
    )
    _write_tle(params, tle_dir)
    stub = _write_stub(params, tmp_dir)
    reader_kwargs = _get_reader_kwargs(params, tle_dir)
    scene = Scene(filenames=[stub], reader="avhrr_l1b_gaclac",
                  reader_kwargs=reader_kwargs)
    scene.load(["1"])
    with pytest.raises(KeyError):
        scene["1"].compute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the satpy.demo module."""
from __future__ import annotations

import contextlib
import io
import os
import sys
import tarfile
import unittest
from collections import defaultdict
from unittest import mock

import pytest

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path
# - tmpdir
# - monkeypatch


class _GlobHelper(object):
    """Create side effect function for mocking gcsfs glob method."""

    def __init__(self, num_results):
        """Initialize side_effect function for mocking gcsfs glob method.

        Args:
            num_results (int or list): Number of results for each glob call
                to return. If a list then number of results per call. The
                last number is used for any additional calls.

        """
        self.current_call = 0
        if not isinstance(num_results, (list, tuple)):
            num_results = [num_results]
        self.num_results = num_results

    def __call__(self, pattern):
        """Mimic glob by being used as the side effect function."""
        try:
            num_results = self.num_results[self.current_call]
        except IndexError:
            num_results = self.num_results[-1]
        self.current_call += 1
        return [pattern + ".{:03d}".format(idx) for idx in range(num_results)]


class TestDemo(unittest.TestCase):
    """Test demo data download functions."""

    def setUp(self):
        """Create temporary directory to save files to."""
        import tempfile
        self.base_dir = tempfile.mkdtemp()
        self.prev_dir = os.getcwd()
        os.chdir(self.base_dir)

    def tearDown(self):
        """Remove the temporary directory created for a test."""
        os.chdir(self.prev_dir)
        try:
            import shutil
            shutil.rmtree(self.base_dir, ignore_errors=True)
        except OSError:
            pass

    @mock.patch("satpy.demo._google_cloud_platform.gcsfs")
    def test_get_us_midlatitude_cyclone_abi(self, gcsfs_mod):
        """Test data download function."""
        from satpy.demo import get_us_midlatitude_cyclone_abi
        gcsfs_mod.GCSFileSystem = mock.MagicMock()
        gcsfs_inst = mock.MagicMock()
        gcsfs_mod.GCSFileSystem.return_value = gcsfs_inst
        gcsfs_inst.glob.return_value = ["a.nc", "b.nc"]
        # expected 16 files, got 2
        with pytest.raises(RuntimeError):
            get_us_midlatitude_cyclone_abi()
        # unknown access method
        with pytest.raises(NotImplementedError):
            get_us_midlatitude_cyclone_abi(method="unknown")

        gcsfs_inst.glob.return_value = ["a.nc"] * 16
        filenames = get_us_midlatitude_cyclone_abi()
        expected = os.path.join(".", "abi_l1b", "20190314_us_midlatitude_cyclone", "a.nc")
        for fn in filenames:
            assert expected == fn

    @mock.patch("satpy.demo._google_cloud_platform.gcsfs")
    def test_get_hurricane_florence_abi(self, gcsfs_mod):
        """Test data download function."""
        from satpy.demo import get_hurricane_florence_abi
        gcsfs_mod.GCSFileSystem = mock.MagicMock()
        gcsfs_inst = mock.MagicMock()
        gcsfs_mod.GCSFileSystem.return_value = gcsfs_inst
        # only return 5 results total
        gcsfs_inst.glob.side_effect = _GlobHelper([5, 0])
        # expected 16 files * 10 frames, got 16 * 5
        with pytest.raises(RuntimeError):
            get_hurricane_florence_abi()
        with pytest.raises(NotImplementedError):
            get_hurricane_florence_abi(method="unknown")

        gcsfs_inst.glob.side_effect = _GlobHelper([int(240 / 16), 0, 0, 0] * 16)
        filenames = get_hurricane_florence_abi()
        assert 10 * 16 == len(filenames)

        gcsfs_inst.glob.side_effect = _GlobHelper([int(240 / 16), 0, 0, 0] * 16)
        filenames = get_hurricane_florence_abi(channels=[2, 3, 4])
        assert 10 * 3 == len(filenames)

        gcsfs_inst.glob.side_effect = _GlobHelper([int(240 / 16), 0, 0, 0] * 16)
        filenames = get_hurricane_florence_abi(channels=[2, 3, 4], num_frames=5)
        assert 5 * 3 == len(filenames)

        gcsfs_inst.glob.side_effect = _GlobHelper([int(240 / 16), 0, 0, 0] * 16)
        filenames = get_hurricane_florence_abi(num_frames=5)
        assert 5 * 16 == len(filenames)


class TestGCPUtils(unittest.TestCase):
    """Test Google Cloud Platform utilities."""

    @mock.patch("satpy.demo._google_cloud_platform.urlopen")
    def test_is_gcp_instance(self, uo):
        """Test is_google_cloud_instance."""
        from satpy.demo._google_cloud_platform import URLError, is_google_cloud_instance
        uo.side_effect = URLError("Test Environment")
        assert not is_google_cloud_instance()

    @mock.patch("satpy.demo._google_cloud_platform.gcsfs")
    def test_get_bucket_files(self, gcsfs_mod):
        """Test get_bucket_files basic cases."""
        from satpy.demo._google_cloud_platform import get_bucket_files
        gcsfs_mod.GCSFileSystem = mock.MagicMock()
        gcsfs_inst = mock.MagicMock()
        gcsfs_mod.GCSFileSystem.return_value = gcsfs_inst
        gcsfs_inst.glob.return_value = ["a.nc", "b.nc"]
        filenames = get_bucket_files("*.nc", ".")
        expected = [os.path.join(".", "a.nc"), os.path.join(".", "b.nc")]
        assert expected == filenames

        gcsfs_inst.glob.side_effect = _GlobHelper(10)
        filenames = get_bucket_files(["*.nc", "*.txt"], ".", pattern_slice=slice(2, 5))
        assert len(filenames) == 3 * 2
        gcsfs_inst.glob.side_effect = None  # reset mock side effect

        gcsfs_inst.glob.return_value = ["a.nc", "b.nc"]
        with pytest.raises(OSError, match="Directory does not exist: does_not_exist"):
            get_bucket_files("*.nc", "does_not_exist")

        open("a.nc", "w").close()  # touch the file
        gcsfs_inst.get.reset_mock()
        gcsfs_inst.glob.return_value = ["a.nc"]
        filenames = get_bucket_files("*.nc", ".")
        assert [os.path.join(".", "a.nc")] == filenames
        gcsfs_inst.get.assert_not_called()

        # force redownload
        gcsfs_inst.get.reset_mock()
        gcsfs_inst.glob.return_value = ["a.nc"]
        filenames = get_bucket_files("*.nc", ".", force=True)
        assert [os.path.join(".", "a.nc")] == filenames
        gcsfs_inst.get.assert_called_once()

        # if we don't get any results then we expect an exception
        gcsfs_inst.get.reset_mock()
        gcsfs_inst.glob.return_value = []
        with pytest.raises(OSError, match="No files could be found or downloaded."):
            get_bucket_files("*.nc", ".")

    @mock.patch("satpy.demo._google_cloud_platform.gcsfs", None)
    def test_no_gcsfs(self):
        """Test that 'gcsfs' is required."""
        from satpy.demo._google_cloud_platform import get_bucket_files
        with pytest.raises(RuntimeError):
            get_bucket_files("*.nc", ".")


class TestAHIDemoDownload:
    """Test the AHI demo data download."""

    @mock.patch.dict(sys.modules, {"s3fs": mock.MagicMock()})
    def test_ahi_full_download(self):
        """Test that the himawari download works as expected."""
        from tempfile import gettempdir

        from satpy.demo import download_typhoon_surigae_ahi
        files = download_typhoon_surigae_ahi(base_dir=gettempdir())
        assert len(files) == 160

    @mock.patch.dict(sys.modules, {"s3fs": mock.MagicMock()})
    def test_ahi_partial_download(self):
        """Test that the himawari download works as expected."""
        from tempfile import gettempdir

        from satpy.demo import download_typhoon_surigae_ahi
        files = download_typhoon_surigae_ahi(base_dir=gettempdir(), segments=[4, 9], channels=[1, 2, 3])
        assert len(files) == 6


def _create_and_populate_dummy_tarfile(fn):
    """Populate a dummy tarfile with dummy files."""
    fn.parent.mkdir(exist_ok=True, parents=True)
    with tarfile.open(fn, mode="x:gz") as tf:
        for i in range(3):
            with open(f"fci-rc{i:d}", "w"):
                pass
            tf.addfile(tf.gettarinfo(name=f"fci-rc{i:d}"))


def test_fci_download(tmp_path, monkeypatch):
    """Test download of FCI test data."""
    from satpy.demo import download_fci_test_data
    monkeypatch.chdir(tmp_path)

    def fake_download_url(url, nm):
        """Create a dummy tarfile.

        Create a dummy tarfile.

        Intended as a drop-in replacement for demo.utils.download_url.
        """
        _create_and_populate_dummy_tarfile(nm)

    with mock.patch("satpy.demo.fci.utils.download_url", new=fake_download_url):
        files = download_fci_test_data(tmp_path)
    assert len(files) == 3
    assert files == ["fci-rc0", "fci-rc1", "fci-rc2"]
    for f in files:
        assert os.path.exists(f)


class _FakeRequest:
    """Fake object to act like a requests return value when downloading a file."""

    requests_log: list[str] = []

    def __init__(self, url, stream=None, timeout=None):
        self._filename = os.path.basename(url)
        self.headers = {}
        self.requests_log.append(url)
        del stream  # just mimicking requests 'get'
        del timeout  # just mimicking requests 'get'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    def raise_for_status(self):
        return

    def _get_fake_bytesio(self):
        filelike_obj = io.BytesIO()
        filelike_obj.write(self._filename.encode("ascii"))
        filelike_obj.seek(0)
        return filelike_obj

    def iter_content(self, chunk_size):
        """Return generator of 'chunk_size' at a time."""
        bytes_io = self._get_fake_bytesio()
        x = bytes_io.read(chunk_size)
        while x:
            yield x
            x = bytes_io.read(chunk_size)


@mock.patch("satpy.demo.utils.requests")
class TestVIIRSSDRDemoDownload:
    """Test VIIRS SDR downloading."""

    ALL_BAND_PREFIXES = ("SVI01", "SVI02", "SVI03", "SVI04", "SVI05",
                         "SVM01", "SVM02", "SVM03", "SVM04", "SVM05", "SVM06", "SVM07", "SVM08", "SVM09", "SVM10",
                         "SVM11", "SVM12", "SVM13", "SVM14", "SVM15", "SVM16",
                         "SVDNB")
    ALL_GEO_PREFIXES = ("GITCO", "GMTCO", "GDNBO")

    def test_download(self, requests, tmpdir):
        """Test downloading VIIRS SDR data."""
        from satpy.demo import get_viirs_sdr_20170128_1229
        requests.get.side_effect = _FakeRequest
        with mock_filesystem():
            files = get_viirs_sdr_20170128_1229(base_dir=str(tmpdir))
            assert len(files) == 10 * (16 + 5 + 1 + 3)  # 10 granules * (5 I bands + 16 M bands + 1 DNB + 3 geolocation)
            self._assert_bands_in_filenames_and_contents(self.ALL_BAND_PREFIXES + self.ALL_GEO_PREFIXES, files, 10)

    def test_do_not_download_the_files_twice(self, requests, tmpdir):
        """Test re-downloading VIIRS SDR data."""
        from satpy.demo import get_viirs_sdr_20170128_1229
        get_mock = mock.MagicMock()
        requests.get.return_value.__enter__ = get_mock
        with mock_filesystem():
            files = get_viirs_sdr_20170128_1229(base_dir=str(tmpdir))
            new_files = get_viirs_sdr_20170128_1229(base_dir=str(tmpdir))

        total_num_files = 10 * (16 + 5 + 1 + 3)  # 10 granules * (5 I bands + 16 M bands + 1 DNB + 3 geolocation)
        assert len(new_files) == total_num_files
        assert get_mock.call_count == total_num_files
        assert new_files == files

    def test_download_channels_num_granules_im(self, requests, tmpdir):
        """Test downloading VIIRS SDR I/M data with select granules."""
        from satpy.demo import get_viirs_sdr_20170128_1229
        requests.get.side_effect = _FakeRequest
        with mock_filesystem():
            files = get_viirs_sdr_20170128_1229(base_dir=str(tmpdir),
                                                channels=("I01", "M01"))
            assert len(files) == 10 * (1 + 1 + 2)  # 10 granules * (1 I band + 1 M band + 2 geolocation)
            self._assert_bands_in_filenames_and_contents(("SVI01", "SVM01", "GITCO", "GMTCO"), files, 10)

    def test_download_channels_num_granules_im_twice(self, requests, tmpdir):
        """Test re-downloading VIIRS SDR I/M data with select granules."""
        from satpy.demo import get_viirs_sdr_20170128_1229
        get_mock = mock.MagicMock()
        requests.get.return_value.__enter__ = get_mock
        with mock_filesystem():
            files = get_viirs_sdr_20170128_1229(base_dir=str(tmpdir),
                                                channels=("I01", "M01"))
            num_first_batch = 10 * (1 + 1 + 2)  # 10 granules * (1 I band + 1 M band + 2 geolocation)
            assert len(files) == num_first_batch

            files = get_viirs_sdr_20170128_1229(base_dir=str(tmpdir),
                                                channels=("I01", "M01"),
                                                granules=(2, 3))
            assert len(files) == 2 * (1 + 1 + 2)  # 2 granules * (1 I band + 1 M band + 2 geolocation)
            assert get_mock.call_count == num_first_batch

    def test_download_channels_num_granules_dnb(self, requests, tmpdir):
        """Test downloading and re-downloading VIIRS SDR DNB data with select granules."""
        from satpy.demo import get_viirs_sdr_20170128_1229
        requests.get.side_effect = _FakeRequest
        with mock_filesystem():
            files = get_viirs_sdr_20170128_1229(base_dir=str(tmpdir),
                                                channels=("DNB",),
                                                granules=(5, 6, 7, 8, 9))
            assert len(files) == 5 * (1 + 1)  # 5 granules * (1 DNB + 1 geolocation)
            self._assert_bands_in_filenames_and_contents(("SVDNB", "GDNBO"), files, 5)

    def _assert_bands_in_filenames_and_contents(self, band_prefixes, filenames, num_files_per_band):
        self._assert_bands_in_filenames(band_prefixes, filenames, num_files_per_band)
        self._assert_file_contents(filenames)

    @staticmethod
    def _assert_bands_in_filenames(band_prefixes, filenames, num_files_per_band):
        for band_name in band_prefixes:
            files_for_band = [x for x in filenames if band_name in x]
            assert files_for_band
            assert len(set(files_for_band)) == num_files_per_band

    @staticmethod
    def _assert_file_contents(filenames):
        for fn in filenames:
            with open(fn, "rb") as fake_hdf5_file:
                assert fake_hdf5_file.read().decode("ascii") == os.path.basename(fn)


@contextlib.contextmanager
def mock_filesystem():
    """Create a mock filesystem, patching `open` and `os.path.isfile`."""
    class FakeFile:
        """Fake file based on BytesIO."""

        def __init__(self):
            self.io = io.BytesIO()

        def __enter__(self):
            return self.io

        def __exit__(self, *args, **kwargs):
            self.io.seek(0)

    fake_fs = defaultdict(FakeFile)
    mo = mock.mock_open()

    def fun(filename, *args, **kwargs):
        return fake_fs[filename]

    mo.side_effect = fun
    with mock.patch("builtins.open", mo):
        with mock.patch("os.path.isfile") as isfile:
            isfile.side_effect = (lambda target: target in fake_fs)
            yield


def test_fs():
    """Test the mock filesystem."""
    with mock_filesystem():
        with open("somefile", "w") as fd:
            fd.write(b"bla")
        with open("someotherfile", "w") as fd:
            fd.write(b"bli")
        with open("somefile", "r") as fd:
            assert fd.read() == b"bla"
        with open("someotherfile", "r") as fd:
            assert fd.read() == b"bli"
        assert os.path.isfile("somefile")
        assert not os.path.isfile("missingfile")


class TestSEVIRIHRITDemoDownload(unittest.TestCase):
    """Test case for downloading an hrit tarball."""

    def setUp(self):
        """Set up the test case."""
        from satpy.demo.seviri_hrit import generate_subset_of_filenames
        self.subdir = os.path.join(".", "seviri_hrit", "20180228_1500")
        self.files = generate_subset_of_filenames(base_dir=self.subdir)

        self.patcher = mock.patch("satpy.demo.utils.requests.get", autospec=True)
        self.get_mock = self.patcher.start()

        _FakeRequest.requests_log = []

    def tearDown(self):
        """Tear down the test case."""
        self.patcher.stop()

    def test_download_gets_files_with_contents(self):
        """Test downloading SEVIRI HRIT data with content."""
        from satpy.demo import download_seviri_hrit_20180228_1500
        self.get_mock.side_effect = _FakeRequest
        with mock_filesystem():
            files = download_seviri_hrit_20180228_1500()
            assert len(files) == 114
            assert set(files) == set(self.files)
            for the_file in files:
                with open(the_file, mode="r") as fd:
                    assert fd.read().decode("utf8") == os.path.basename(the_file)

    def test_download_from_zenodo(self):
        """Test downloading SEVIRI HRIT data from zenodo."""
        from satpy.demo import download_seviri_hrit_20180228_1500
        self.get_mock.side_effect = _FakeRequest
        with mock_filesystem():
            download_seviri_hrit_20180228_1500()
            assert _FakeRequest.requests_log[0].startswith("https://zenodo.org")

    def test_download_a_subset_of_files(self):
        """Test downloading a subset of files."""
        from satpy.demo import download_seviri_hrit_20180228_1500
        with mock_filesystem():
            files = download_seviri_hrit_20180228_1500(subset={"HRV": [1, 2, 3], "IR_108": [1, 2], "EPI": None})
            assert set(files) == set(os.path.join(self.subdir, filename) for filename in [
                "H-000-MSG4__-MSG4________-_________-EPI______-201802281500-__",
                "H-000-MSG4__-MSG4________-HRV______-000001___-201802281500-__",
                "H-000-MSG4__-MSG4________-HRV______-000002___-201802281500-__",
                "H-000-MSG4__-MSG4________-HRV______-000003___-201802281500-__",
                "H-000-MSG4__-MSG4________-IR_108___-000001___-201802281500-__",
                "H-000-MSG4__-MSG4________-IR_108___-000002___-201802281500-__",
            ])

    def test_do_not_download_same_file_twice(self):
        """Test that files are not downloaded twice."""
        from satpy.demo import download_seviri_hrit_20180228_1500
        get_mock = mock.MagicMock()
        self.get_mock.return_value.__enter__ = get_mock
        with mock_filesystem():
            files = download_seviri_hrit_20180228_1500(subset={"HRV": [1, 2, 3], "IR_108": [1, 2], "EPI": None})
            new_files = download_seviri_hrit_20180228_1500(subset={"HRV": [1, 2, 3], "IR_108": [1, 2], "EPI": None})
            assert set(files) == set(new_files)
            assert get_mock.call_count == 6

    def test_download_to_output_directory(self):
        """Test downloading to an output directory."""
        from tempfile import gettempdir

        from satpy.demo import download_seviri_hrit_20180228_1500
        with mock_filesystem():
            base_dir = gettempdir()
            files = download_seviri_hrit_20180228_1500(base_dir=base_dir)
            assert files[0].startswith(base_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utilties for getting various angles for a dataset.."""
from __future__ import annotations

import datetime as dt
import hashlib
import os
import shutil
import warnings
from functools import update_wrapper
from glob import glob
from typing import Any, Callable, Optional, Union

import dask
import numpy as np
import xarray as xr
from dask import array as da
from pyorbital.astronomy import cos_zen as pyob_cos_zen
from pyorbital.astronomy import get_alt_az
from pyorbital.orbital import get_observer_look
from pyresample.geometry import AreaDefinition, StackedAreaDefinition, SwathDefinition

import satpy
from satpy.utils import PerformanceWarning, get_satpos, ignore_invalid_float_warnings

PRGeometry = Union[SwathDefinition, AreaDefinition, StackedAreaDefinition]

# Arbitrary time used when computing sensor angles that is passed to
# pyorbital's get_observer_look function.
# The difference is on the order of 1e-10 at most as time changes so we force
# it to a single time for easier caching. It is *only* used if caching.
STATIC_EARTH_INERTIAL_DATETIME = dt.datetime(2000, 1, 1, 12, 0, 0)
DEFAULT_UNCACHE_TYPES = (SwathDefinition, xr.DataArray, da.Array)
HASHABLE_GEOMETRIES = (AreaDefinition, StackedAreaDefinition)


class ZarrCacheHelper:
    """Helper for caching function results to on-disk zarr arrays.

    It is recommended to use this class through the :func:`cache_to_zarr_if`
    decorator rather than using it directly.

    Currently the cache does not perform any limiting or removal of cache
    content. That is left up to the user to manage. Caching is based on
    arguments passed to the decorated function but will only be performed
    if the arguments are of a certain type (see ``uncacheable_arg_types``).
    The cache value to use is purely based on the hash value of all of the
    provided arguments along with the "cache version" (see below).

    Note that the zarr format requires regular chunking of data. That is,
    chunks must be all the same size per dimension except for the last chunk.
    To work around this limitation, this class will determine a good regular
    chunking based on the existing chunking scheme, rechunk the input
    arguments, and then rechunk the results before returning them to the user.
    This rechunking is only done if caching is enabled.

    Args:
        func: Function that will be called to generate the value to cache.
        cache_config_key: Name of the boolean ``satpy.config`` parameter to
            use to determine if caching should be done.
        uncacheable_arg_types: Types that if present in the passed arguments
            should trigger caching to *not* happen. By default this includes
            ``SwathDefinition``, ``xr.DataArray``, and ``dask.array.Array`` objects.
        sanitize_args_func: Optional function to call to sanitize provided
            arguments before they are considered for caching. This can be used
            to make arguments more "cacheable" by replacing them with similar
            values that will result in more cache hits. Note that the sanitized
            arguments are only passed to the underlying function if caching
            will be performed, otherwise the original arguments are passed.
        cache_version: Version number used to distinguish one version of a
            decorated function from future versions.

    Notes:
        * Caching only supports dask array values.

        * This helper allows for an additional ``cache_dir`` parameter to
          override the use of the ``satpy.config`` ``cache_dir`` parameter.

    Examples:
        To use through the :func:`cache_to_zarr_if` decorator::

            @cache_to_zarr_if("cache_my_stuff")
            def generate_my_stuff(area_def: AreaDefinition, some_factor: int) -> da.Array:
                # Generate
                return my_dask_arr

        To use the decorated function::

            with satpy.config.set(cache_my_stuff=True):
                my_stuff = generate_my_stuff(area_def, 5)

    """

    def __init__(self,
                 func: Callable,
                 cache_config_key: str,
                 uncacheable_arg_types=DEFAULT_UNCACHE_TYPES,
                 sanitize_args_func: Optional[Callable] = None,
                 cache_version: int = 1,
                 ):
        """Hold on to provided arguments for future use."""
        self._func = func
        self._cache_config_key = cache_config_key
        self._uncacheable_arg_types = uncacheable_arg_types
        self._sanitize_args_func = sanitize_args_func
        self._cache_version = cache_version

    def cache_clear(self, cache_dir: Optional[str] = None):
        """Remove all on-disk files associated with this function.

        Intended to mimic the :func:`functools.cache` behavior.
        """
        cache_dir = self._get_cache_dir_from_config(cache_dir)
        zarr_pattern = self._zarr_pattern("*", cache_version="*").format("*")
        for zarr_dir in glob(os.path.join(cache_dir, zarr_pattern)):
            shutil.rmtree(zarr_dir, ignore_errors=True)

    def _zarr_pattern(self, arg_hash, cache_version: Union[None, int, str] = None) -> str:
        if cache_version is None:
            cache_version = self._cache_version
        return f"{self._func.__name__}_v{cache_version}" + "_{}_" + f"{arg_hash}.zarr"

    def __call__(self, *args, cache_dir: Optional[str] = None) -> Any:
        """Call the decorated function."""
        should_cache: bool = satpy.config.get(self._cache_config_key, False)
        if not should_cache:
            return self._func(*args)

        try:
            return self._cache_and_read(args, cache_dir)
        except TypeError as err:
            warnings.warn("Cannot cache function because of unhashable argument: " + str(err), stacklevel=2)
            return self._func(*args)

    def _cache_and_read(self, args, cache_dir):
        sanitized_args = self._sanitize_args_func(*args) if self._sanitize_args_func is not None else args

        zarr_file_pattern = self._get_zarr_file_pattern(sanitized_args, cache_dir)
        zarr_paths = glob(zarr_file_pattern.format("*"))

        if not zarr_paths:
            # use sanitized arguments
            self._warn_if_irregular_input_chunks(args, sanitized_args)
            res_to_cache = self._func(*(sanitized_args))
            self._cache_results(res_to_cache, zarr_file_pattern)

        # if we did any caching, let's load from the zarr files, so that future calls have the same name
        # re-calculate the cached paths
        zarr_paths = sorted(glob(zarr_file_pattern.format("*")))
        if not zarr_paths:
            raise RuntimeError("Data was cached to disk but no files were found")

        new_chunks = _get_output_chunks_from_func_arguments(args)
        res = tuple(da.from_zarr(zarr_path, chunks=new_chunks) for zarr_path in zarr_paths)
        return res

    def _get_zarr_file_pattern(self, sanitized_args, cache_dir):
        arg_hash = _hash_args(*sanitized_args, unhashable_types=self._uncacheable_arg_types)
        zarr_filename = self._zarr_pattern(arg_hash)
        cache_dir = self._get_cache_dir_from_config(cache_dir)
        return os.path.join(cache_dir, zarr_filename)

    @staticmethod
    def _get_cache_dir_from_config(cache_dir: Optional[str]) -> str:
        cache_dir = cache_dir or satpy.config.get("cache_dir")
        if cache_dir is None:
            raise RuntimeError("Can't use zarr caching. No 'cache_dir' configured.")
        return cache_dir

    @staticmethod
    def _warn_if_irregular_input_chunks(args, modified_args):
        arg_chunks = _get_output_chunks_from_func_arguments(args)
        new_chunks = _get_output_chunks_from_func_arguments(modified_args)
        if _chunks_are_irregular(arg_chunks):
            warnings.warn(
                "Calling cached function with irregular dask chunks. The data "
                "has been rechunked for caching, but this is not optimal for "
                "future calculations. "
                f"Original chunks: {arg_chunks}; New chunks: {new_chunks}",
                PerformanceWarning,
                stacklevel=3
            )

    def _cache_results(self, res, zarr_file_pattern):
        os.makedirs(os.path.dirname(zarr_file_pattern), exist_ok=True)
        new_res = []
        for idx, sub_res in enumerate(res):
            if not isinstance(sub_res, da.Array):
                raise ValueError("Zarr caching currently only supports dask "
                                 f"arrays. Got {type(sub_res)}")
            zarr_path = zarr_file_pattern.format(idx)
            # See https://github.com/dask/dask/issues/8380
            with dask.config.set({"optimization.fuse.active": False}):
                new_sub_res = sub_res.to_zarr(zarr_path, compute=False)
            new_res.append(new_sub_res)
        # actually compute the storage to zarr
        da.compute(new_res)


def _get_output_chunks_from_func_arguments(args):
    """Determine what the desired output chunks are.

    It is assumed a tuple of tuples of integers is defining chunk sizes. If
    a tuple like this is not found then arguments are checked for array-like
    objects with a ``.chunks`` attribute.

    """
    chunked_args = [arg for arg in args if hasattr(arg, "chunks")]
    tuple_args = [arg for arg in args if _is_chunk_tuple(arg)]
    if not tuple_args and not chunked_args:
        raise RuntimeError("Cannot determine desired output chunksize for cached function.")
    new_chunks = tuple_args[-1] if tuple_args else chunked_args[0].chunks
    return new_chunks


def cache_to_zarr_if(
        cache_config_key: str,
        uncacheable_arg_types=DEFAULT_UNCACHE_TYPES,
        sanitize_args_func: Optional[Callable] = None,
) -> Callable:
    """Decorate a function and cache the results as a zarr array on disk.

    This only happens if the ``satpy.config`` boolean value for the provided
    key is ``True`` as well as some other conditions. See
    :class:`ZarrCacheHelper` for more information. Most importantly, this
    decorator does not limit how many items can be cached and does not clear
    out old entries. It is up to the user to manage the size of the cache.

    """

    def _decorator(func: Callable) -> Callable:
        zarr_cacher = ZarrCacheHelper(func,
                                      cache_config_key,
                                      uncacheable_arg_types,
                                      sanitize_args_func)
        wrapper = update_wrapper(zarr_cacher, func)
        return wrapper

    return _decorator


def _hash_args(*args, unhashable_types=DEFAULT_UNCACHE_TYPES):
    import json
    hashable_args = []
    for arg in args:
        if isinstance(arg, unhashable_types):
            raise TypeError(f"Unhashable type ({type(arg)}).")
        if isinstance(arg, HASHABLE_GEOMETRIES):
            arg = hash(arg)
        elif isinstance(arg, dt.datetime):
            arg = arg.isoformat(" ")
        hashable_args.append(arg)
    arg_hash = hashlib.sha1()  # nosec
    arg_hash.update(json.dumps(tuple(hashable_args)).encode("utf8"))
    return arg_hash.hexdigest()


def _sanitize_observer_look_args(*args):
    new_args = []
    for arg in args:
        if isinstance(arg, dt.datetime):
            new_args.append(STATIC_EARTH_INERTIAL_DATETIME)
        elif isinstance(arg, (float, np.float64, np.float32)):
            # Round floating point numbers to nearest tenth. Numpy types don't
            # serialize into JSON which is needed for hashing, thus the casting
            # to float here:
            new_args.append(float(round(arg, 1)))
        elif _is_chunk_tuple(arg) and _chunks_are_irregular(arg):
            new_chunks = _regular_chunks_from_irregular_chunks(arg)
            new_args.append(new_chunks)
        else:
            new_args.append(arg)
    return new_args


def _sanitize_args_with_chunks(*args):
    new_args = []
    for arg in args:
        if _is_chunk_tuple(arg) and _chunks_are_irregular(arg):
            new_chunks = _regular_chunks_from_irregular_chunks(arg)
            new_args.append(new_chunks)
        else:
            new_args.append(arg)
    return new_args


def _is_chunk_tuple(some_obj: Any) -> bool:
    if not isinstance(some_obj, tuple):
        return False
    if not all(isinstance(sub_obj, tuple) for sub_obj in some_obj):
        return False
    sub_elements = [sub_obj_elem for sub_obj in some_obj for sub_obj_elem in sub_obj]
    return all(isinstance(sub_obj_elem, int) for sub_obj_elem in sub_elements)


def _regular_chunks_from_irregular_chunks(
        old_chunks: tuple[tuple[int, ...], ...]
) -> tuple[tuple[int, ...], ...]:
    shape = tuple(sum(dim_chunks) for dim_chunks in old_chunks)
    new_dim_chunks = tuple(max(dim_chunks) for dim_chunks in old_chunks)
    return da.core.normalize_chunks(new_dim_chunks, shape=shape)


def _chunks_are_irregular(chunks_tuple: tuple) -> bool:
    """Determine if an array is irregularly chunked.

    Zarr does not support saving data in irregular chunks. Regular chunking
    is when all chunks are the same size (except for the last one).

    """
    if any(len(set(chunks[:-1])) > 1 for chunks in chunks_tuple):
        return True
    return any(chunks[-1] > chunks[0] for chunks in chunks_tuple)


def _geo_dask_to_data_array(arr: da.Array) -> xr.DataArray:
    return xr.DataArray(arr, dims=("y", "x"))


def compute_relative_azimuth(
        sat_azi: xr.DataArray | da.Array,
        sun_azi: xr.DataArray | da.Array
) -> xr.DataArray | da.Array:
    """Compute the relative azimuth angle.

    Args:
        sat_azi: satellite azimuth angles typically in the 0-360 degree range.
        sun_azi: solar azimuth angles in same range as sat_azi.

    Returns:
        The relative azimuth angle or difference between solar and satellite
        azimuth angles in the 0-180 degree range.

    NOTE: Relative azimuth is defined such that:
    Relative azimuth is 0 when sun and satellite are aligned on one side of a pixel (back scatter).
    Relative azimuth is 180 when sun and satellite are directly opposite each other (forward scatter).
    """
    xarray_dims = getattr(sat_azi, "dims", None)
    xarray_coords = getattr(sat_azi, "coords", None)
    if xarray_dims is not None:
        sat_azi = sat_azi.data
        sun_azi = sun_azi.data
    rel_azi = da.map_blocks(
        _compute_relative_azimuth,
        sat_azi, sun_azi,
        dtype=sat_azi.dtype,
        meta=np.array((), dtype=sat_azi.dtype),
        token="relative_azimuth",  # nosec: B106
    )
    if xarray_dims is None:
        return rel_azi
    return xr.DataArray(rel_azi, dims=xarray_dims, coords=xarray_coords)


def _compute_relative_azimuth(sat_azi: np.ndarray, sun_azi: np.ndarray) -> np.ndarray:
    ssadiff = np.absolute(sun_azi - sat_azi)
    dtype = sun_azi.dtype.type
    return np.minimum(ssadiff, dtype(360.0) - ssadiff)


def get_angles(data_arr: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray, xr.DataArray, xr.DataArray]:
    """Get sun and satellite azimuth and zenith angles.

    Note that this function can benefit from the ``satpy.config`` parameters
    :ref:`cache_lonlats <config_cache_lonlats_setting>` and
    :ref:`cache_sensor_angles <config_cache_sensor_angles_setting>`
    being set to ``True``.

    Args:
        data_arr: DataArray to get angles for. Information extracted from this
            object are ``.attrs["area"]``,``.attrs["start_time"]``, and
            ``.attrs["orbital_parameters"]``. See :func:`satpy.utils.get_satpos`
            and :ref:`dataset_metadata` for more information.
            Additionally, the dask array chunk size is used when generating
            new arrays. The actual data of the object is not used.

    Returns:
        Four DataArrays representing sensor azimuth angle, sensor zenith angle,
        solar azimuth angle, and solar zenith angle. All values are in degrees.
        Sensor angles are provided in the [0, 360] degree range.
        Solar angles are provided in the [-180, 180] degree range.

    """
    sata, satz = _get_sensor_angles(data_arr)
    suna, sunz = _get_sun_angles(data_arr)
    return sata, satz, suna, sunz


def get_satellite_zenith_angle(data_arr: xr.DataArray) -> xr.DataArray:
    """Generate satellite zenith angle for the provided data.

    Note that this function can benefit from the ``satpy.config`` parameters
    :ref:`cache_lonlats <config_cache_lonlats_setting>` and
    :ref:`cache_sensor_angles <config_cache_sensor_angles_setting>`
    being set to ``True``. Values are in degrees.

    """
    satz = _get_sensor_angles(data_arr)[1]
    return satz


def get_cos_sza(data_arr: xr.DataArray) -> xr.DataArray:
    """Generate the cosine of the solar zenith angle for the provided data.

    Returns:
        DataArray with the same shape as ``data_arr``.

    """
    chunks = _geo_chunks_from_data_arr(data_arr)
    lons, lats = _get_valid_lonlats(data_arr.attrs["area"], chunks)
    if lons.dtype != data_arr.dtype and np.issubdtype(data_arr.dtype, np.floating):
        lons = lons.astype(data_arr.dtype)
        lats = lats.astype(data_arr.dtype)
    cos_sza = _get_cos_sza(data_arr.attrs["start_time"], lons, lats)
    return _geo_dask_to_data_array(cos_sza)


@cache_to_zarr_if("cache_lonlats", sanitize_args_func=_sanitize_args_with_chunks)
def _get_valid_lonlats(area: PRGeometry, chunks: Union[int, str, tuple] = "auto") -> tuple[da.Array, da.Array]:
    with ignore_invalid_float_warnings():
        # NOTE: This defaults to 64-bit floats due to needed precision for X/Y coordinates
        lons, lats = area.get_lonlats(chunks=chunks)
        lons = da.where(lons >= 1e30, np.nan, lons)
        lats = da.where(lats >= 1e30, np.nan, lats)
    return lons, lats


def _get_sun_angles(data_arr: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray]:
    chunks = _geo_chunks_from_data_arr(data_arr)
    lons, lats = _get_valid_lonlats(data_arr.attrs["area"], chunks)
    suna = da.map_blocks(_get_sun_azimuth_ndarray, lons, lats,
                         data_arr.attrs["start_time"],
                         dtype=lons.dtype, meta=np.array((), dtype=lons.dtype),
                         chunks=lons.chunks)
    cos_sza = _get_cos_sza(data_arr.attrs["start_time"], lons, lats)
    sunz = np.rad2deg(np.arccos(cos_sza))
    suna = _geo_dask_to_data_array(suna)
    sunz = _geo_dask_to_data_array(sunz)
    return suna, sunz


def _get_cos_sza(utc_time, lons, lats):
    cos_sza = da.map_blocks(_cos_zen_ndarray,
                            lons, lats, utc_time,
                            meta=np.array((), dtype=lons.dtype),
                            dtype=lons.dtype,
                            chunks=lons.chunks)
    return cos_sza


def _cos_zen_ndarray(lons, lats, utc_time):
    with ignore_invalid_float_warnings():
        return pyob_cos_zen(utc_time, lons, lats)


def _get_sun_azimuth_ndarray(lons: np.ndarray, lats: np.ndarray, start_time: dt.datetime) -> np.ndarray:
    with ignore_invalid_float_warnings():
        suna = get_alt_az(start_time, lons, lats)[1]
        suna = np.rad2deg(suna)

        # The get_alt_az function returns values in the range -180 to 180 degrees.
        # Satpy expects values in the 0 - 360 range, which is what is returned for the
        # satellite azimuth angles.
        # Here this is corrected so both sun and sat azimuths are in the same range.
        suna = suna % 360.
    return suna


def _get_sensor_angles(data_arr: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray]:
    preference = satpy.config.get("sensor_angles_position_preference", "actual")
    sat_lon, sat_lat, sat_alt = get_satpos(data_arr, preference=preference)
    area_def = data_arr.attrs["area"]
    chunks = _geo_chunks_from_data_arr(data_arr)

    sata, satz = _get_sensor_angles_from_sat_pos(sat_lon, sat_lat, sat_alt,
                                                 data_arr.attrs["start_time"],
                                                 area_def, chunks)
    sata = _geo_dask_to_data_array(sata)
    satz = _geo_dask_to_data_array(satz)
    return sata, satz


def _geo_chunks_from_data_arr(data_arr: xr.DataArray) -> tuple:
    x_dim_index = _dim_index_with_default(data_arr.dims, "x", -1)
    y_dim_index = _dim_index_with_default(data_arr.dims, "y", -2)
    chunks = (data_arr.chunks[y_dim_index], data_arr.chunks[x_dim_index])
    return chunks


def _dim_index_with_default(dims: tuple, dim_name: str, default: int) -> int:
    try:
        return dims.index(dim_name)
    except ValueError:
        return default


@cache_to_zarr_if("cache_sensor_angles", sanitize_args_func=_sanitize_observer_look_args)
def _get_sensor_angles_from_sat_pos(sat_lon, sat_lat, sat_alt, start_time, area_def, chunks):
    lons, lats = _get_valid_lonlats(area_def, chunks)
    res = da.map_blocks(_get_sensor_angles_ndarray, lons, lats, start_time, sat_lon, sat_lat, sat_alt,
                        dtype=lons.dtype, meta=np.array((), dtype=lons.dtype), new_axis=[0],
                        chunks=(2,) + lons.chunks)
    return res[0], res[1]


def _get_sensor_angles_ndarray(lons, lats, start_time, sat_lon, sat_lat, sat_alt) -> np.ndarray:
    with ignore_invalid_float_warnings():
        sata, satel = get_observer_look(
            sat_lon,
            sat_lat,
            sat_alt / 1000.0,  # km
            start_time,
            lons, lats, 0)
        satz = 90 - satel
        return np.stack([sata, satz])


def sunzen_corr_cos(data: da.Array,
                    cos_zen: da.Array,
                    limit: float = 88.,
                    max_sza: Optional[float] = 95.) -> da.Array:
    """Perform Sun zenith angle correction.

    The correction is based on the provided cosine of the zenith
    angle (``cos_zen``).  The correction is limited
    to ``limit`` degrees (default: 88.0 degrees).  For larger zenith
    angles, the correction is the same as at the ``limit`` if ``max_sza``
    is `None`. The default behavior is to gradually reduce the correction
    past ``limit`` degrees up to ``max_sza`` where the correction becomes
    0. Both ``data`` and ``cos_zen`` should be 2D arrays of the same shape.

    """
    return da.map_blocks(_sunzen_corr_cos_ndarray,
                         data, cos_zen, limit, max_sza,
                         meta=np.array((), dtype=data.dtype),
                         chunks=data.chunks)


def _sunzen_corr_cos_ndarray(data: np.ndarray,
                             cos_zen: np.ndarray,
                             limit: float,
                             max_sza: Optional[float]) -> np.ndarray:
    # Convert the zenith angle limit to cosine of zenith angle
    limit_rad = np.deg2rad(limit)
    limit_cos = np.cos(limit_rad)
    max_sza_rad = np.deg2rad(max_sza) if max_sza is not None else max_sza

    # Cosine correction
    corr = (1. / cos_zen).astype(data.dtype, copy=False)
    if max_sza is not None:
        # gradually fall off for larger zenith angle
        grad_factor = (np.arccos(cos_zen) - limit_rad) / (max_sza_rad - limit_rad)
        # invert the factor so maximum correction is done at `limit` and falls off later
        with np.errstate(invalid="ignore"):  # we expect space pixels to be invalid
            grad_factor = 1. - np.log(grad_factor + 1) / np.log(2)
        # make sure we don't make anything negative
        grad_factor = grad_factor.clip(0.)
    else:
        # Use constant value (the limit) for larger zenith angles
        grad_factor = 1.
    corr = np.where(
        cos_zen > limit_cos,
        corr,
        (grad_factor / limit_cos).astype(data.dtype, copy=False)
    )
    # Force "night" pixels to 0 (where SZA is invalid)
    corr[np.isnan(cos_zen)] = 0
    return data * corr


def sunzen_reduction(data: da.Array,
                     sunz: da.Array,
                     limit: float = 55.,
                     max_sza: float = 90.,
                     strength: float = 1.5) -> da.Array:
    """Reduced strength of signal at high sun zenith angles."""
    return da.map_blocks(_sunzen_reduction_ndarray, data, sunz, limit, max_sza, strength,
                         meta=np.array((), dtype=data.dtype), chunks=data.chunks)

def _sunzen_reduction_ndarray(data: np.ndarray,
                              sunz: np.ndarray,
                              limit: float,
                              max_sza: float,
                              strength: float) -> np.ndarray:
    # compute reduction factor (0.0 - 1.0) between limit and maz_sza
    reduction_factor = (sunz - limit) / (max_sza - limit)
    reduction_factor = reduction_factor.clip(0., 1.)

    # invert the reduction factor such that minimum reduction is done at `limit` and gradually increases towards max_sza
    with np.errstate(invalid="ignore"):  # we expect space pixels to be invalid
        reduction_factor = 1. - np.log2(reduction_factor + 1)

    # apply non-linearity to the reduction factor for a non-linear reduction of the signal. This can be used for a
    # slower or faster transision to higher/lower fractions at the ndvi extremes. If strength equals 1.0, this
    # operation has no effect on the reduction_factor.
    reduction_factor = reduction_factor ** strength / (
                reduction_factor ** strength + (1 - reduction_factor) ** strength)

    # compute final correction term, with no reduction for angles < limit
    corr = np.where(sunz < limit, 1.0, reduction_factor)

    # force "night" pixels to 0 (where SZA is invalid)
    corr[np.isnan(sunz)] = 0

    # reduce data signal with correction term
    res = data * corr

    return res
